from deepchem.feat import Featurizer
from typing import List
import numpy as np

try:
    from transformers import RobertaTokenizerFast
except ModuleNotFoundError:
    raise ImportError(
        'Transformers must be installed for RxnFeaturizer to be used!')
    pass


class RxnFeaturizer(Featurizer):
    """Reaction Featurizer.

    RxnFeaturizer is a wrapper class for HuggingFace's RobertaTokenizerFast,
    that is intended for featurizing chemical reaction datasets. The featurizer
    computes the source and target required for a seq2seq task and applies the
    RobertaTokenizer on them separately. Additionally, it can also separate or
    mix the reactants and reagents before tokenizing.

    Examples
    --------
    >>> from deepchem.feat import RxnFeaturizer
    >>> from transformers import RobertaTokenizerFast
    >>> tokenizer = RobertaTokenizerFast.from_pretrained("seyonec/PubChem10M_SMILES_BPE_450k")
    >>> featurizer = RxnFeaturizer(tokenizer, sep_reagent=True)
    >>> feats = featurizer.featurize(['CCS(=O)(=O)Cl.OCCBr>CCN(CC)CC.CCOCC>CCS(=O)(=O)OCCBr'])

    Notes
    -----
    - The featurize method expects a List of reactions.
    - Use the sep_reagent toggle to enable/disable reagent separation.
        - True - Separate the reactants and reagents
        - False - Mix the reactants and reagents
    """

    def __init__(self, tokenizer: RobertaTokenizerFast, sep_reagent: bool):
        """Initialize a ReactionFeaturizer object.

        Parameters
        ----------
        tokenizer: RobertaTokenizerFast
            HuggingFace Tokenizer to be used for featurization.
        sep_reagent: bool
            Toggle to separate or mix the reactants and reagents.
        """
        if not isinstance(tokenizer, RobertaTokenizerFast):
            raise TypeError(
                f"""`tokenizer` must be a constructed `RobertaTokenizerFast`
                object, not {type(tokenizer)}""")
        else:
            self.tokenizer = tokenizer
        self.sep_reagent = sep_reagent

    def _featurize(self, datapoint: str, **kwargs) -> List[List[List[int]]]:
        """Featurizes a datapoint.

        Processes each entry in the dataset by first applying the reactant-reagent
        mixing, the source/target separation and then the pretrained tokenizer on the
        separated strings.

        Parameters
        ----------
        datapoint: str
            the reaction SMILES to be processed.

        Returns
        -------
        encoding: List
            List containing two lists for the source and target encodings.
            The encodings are lists containing two lists: `the input_ids` and the
        `attention_mask`.
        """

        datapoint_list = [datapoint]
        reactant = list(map(lambda x: x.split('>')[0], datapoint_list))
        reagent = list(map(lambda x: x.split('>')[1], datapoint_list))
        product = list(map(lambda x: x.split('>')[2], datapoint_list))

        if self.sep_reagent:
            source = [x + '>' + y for x, y in zip(reactant, reagent)]
        else:
            source = [
                x + '.' + y + '>' if y else x + '>' + y
                for x, y in zip(reactant, reagent)
            ]
        target = product

        source_encoding = list(
            self.tokenizer(source, padding=True, **kwargs).values())
        target_encoding = list(
            self.tokenizer(target, padding=True, **kwargs).values())

        return [source_encoding, target_encoding]

    def __call__(self, *args, **kwargs) -> np.ndarray:
        return self.featurize(*args, **kwargs)

    def __str__(self) -> str:
        """Handles file name error.

        Overrides the __str__ method of the Featurizer base class to avoid errors
        while saving the dataset, due to the large default name of the HuggingFace
        tokenizer.
        """
        return 'RxnFeaturizer_' + str(self.sep_reagent)



"""
SWEET dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

SWEETLEAD_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/sweet.csv.gz"
SWEETLEAD_TASKS = ["task"]


class _SweetLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "sweet.csv.gz")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=SWEETLEAD_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_sweet(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['balancing'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load sweet datasets.

    Sweetlead is a dataset of chemical structures for approved drugs, chemical isolates
    from traditional medicinal herbs, and regulated chemicals. Resulting structures are
    filtered for the active pharmaceutical ingredient, standardized, and differing
    formulations of the same drug were combined in the final database.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    References
    ----------
    Novick, Paul A., et al. "SWEETLEAD: an in silico database of approved drugs, regulated
    chemicals, and herbal isolates for computer-aided drug discovery." PLoS One 8.11 (2013).
    """
    loader = _SweetLoader(featurizer, splitter, transformers, SWEETLEAD_TASKS,
                          data_dir, save_dir, **kwargs)
    return loader.load_dataset('sweet', reload)



import numpy as np
from typing import Optional, Callable


class ElectronSampler:
    """
    This class enables to initialize electron's position using gauss distribution around a nucleus and update using Markov Chain Monte-Carlo(MCMC) moves.

    Using the probability obtained from the square of magnitude of wavefunction of a molecule/atom, MCMC steps can be performed to get the electron's positions and further update the wavefunction.
    This method is primarily used in methods like Variational Monte Carlo to sample electrons around the nucleons.
    Sampling can be done in 2 ways:
    -Simultaneous: All the electrons' positions are updated all at once.

    -Single-electron: MCMC steps are performed only a particular electron, given their index value.

    Further these moves can be done in 2 methods:
    -Symmetric: In this configuration, the standard deviation for all the steps are uniform.

    -Asymmetric: In this configuration, the standard deviation are not uniform and typically the standard deviation is obtained a function like harmonic distances, etc.

    Irrespective of these methods, the initialization is done uniformly around the respective nucleus and the number of electrons specified.

    Example
    -------
    >>> from deepchem.utils.electron_sampler import ElectronSampler
    >>> test_f = lambda x: 2*np.log(np.random.uniform(low=0,high=1.0,size=np.shape(x)[0]))
    >>> distribution=ElectronSampler(central_value=np.array([[1,1,3],[3,2,3]]),f=test_f,seed=0,batch_no=2,steps=1000,)
    >>> distribution.gauss_initialize_position(np.array([[1],[2]]))

    >> print(distribution.x)
    [[[[1.03528105 1.00800314 3.01957476]]

      [[3.01900177 1.99697286 2.99793562]]

      [[3.00821197 2.00288087 3.02908547]]]


    [[[1.04481786 1.03735116 2.98045444]]

      [[3.01522075 2.0024335  3.00887726]]

      [[3.00667349 2.02988158 2.99589683]]]]
    >>> distribution.move()
    0.5115

    >> print(distribution.x)
    [[[[-0.32441754  1.23330263  2.67927645]]

      [[ 3.42250997  2.23617126  3.55806632]]

      [[ 3.37491385  1.54374006  3.13575241]]]


    [[[ 0.49067726  1.03987841  3.70277884]]

      [[ 3.5631939   1.68703947  2.5685874 ]]

      [[ 2.84560249  1.73998364  3.41274181]]]]
    """

    def __init__(self,
                 central_value: np.ndarray,
                 f: Callable[[np.ndarray], np.ndarray],
                 batch_no: int = 10,
                 x: np.ndarray = np.array([]),
                 steps: int = 10,
                 seed: Optional[int] = None,
                 symmetric: bool = True,
                 simultaneous: bool = True):
        """
        Parameters:
        -----------
        central_value: np.ndarray
            Contains each nucleus' coordinates in a 2D array. The shape of the array should be(number_of_nucleus,3).Ex: [[1,2,3],[3,4,5],..]
        f:Callable[[np.ndarray],np.ndarray]
            A function that should give the twice the log probability of wavefunction of the molecular system when called. Should taken in a 4D array of electron's positions(x) as argument and return a numpy array containing the log probabilities of each batch.
        batch_no: int, optional (default 10)
            Number of batches of the electron's positions to be initialized.
        x: np.ndarray, optional (default np.ndarray([]))
            Contains the electron's coordinates in a 4D array. The shape of the array should be(batch_no,no_of_electrons,1,3). Can be a 1D empty array, when electron's positions are yet to be initialized.
        steps: int, optional (default 10)
            The number of MCMC steps to be performed when the moves are called.
        seed: int, optional (default None)
            Random seed to use.
        symmetric: bool, optional(default True)
            If true, symmetric moves will be used, else asymmetric moves will be followed.
        simultaneous: bool, optional(default True)
            If true, MCMC steps will be performed on all the electrons, else only a single electron gets updated.
        """
        self.x = x
        self.f = f
        self.num_accept = 0
        self.symmetric = symmetric
        self.simultaneous = simultaneous
        self.steps = steps
        self.central_value = central_value
        self.batch_no = batch_no
        if seed is not None:
            seed = int(seed)
            np.random.seed(seed)

    def harmonic_mean(self, y: np.ndarray) -> np.ndarray:
        """Calculates the harmonic mean of the value 'y' from the self.central value. The numpy array returned is typically scaled up to get the standard deviation matrix.

        Parameters
        ----------
        y: np.ndarray
            Containing the data distribution. Shape of y should be (batch,no_of_electron,1,3)

        Return
        ----------
        np.ndarray
            Contains the harmonic mean of the data distribution of each batch. Shape of the array obtained (batch_no, no_of_electrons,1,1)
        """

        diff = y - self.central_value
        distance = np.linalg.norm(diff, axis=-1, keepdims=True)
        return 1.0 / np.mean(1.0 / distance, axis=-2, keepdims=True)

    def log_prob_gaussian(self, y: np.ndarray, mu: np.ndarray,
                          sigma: np.ndarray) -> np.ndarray:
        """Calculates the log probability of a gaussian distribution, given the mean and standard deviation

        Parameters
        ----------
        y: np.ndarray
            data for which the log normal distribution is to be found
        mu: np.ndarray
            Means wrt which the log normal is calculated. Same shape as x or should be brodcastable to x
        sigma: np.ndarray,
            The standard deviation of the log normal distribution. Same shape as x or should be brodcastable to x

        Return
        ----------
        np.ndarray
            Log probability of gaussian distribution, with the shape - (batch_no,).
        """

        numer = np.sum((-0.5 * ((y - mu)**2) / (sigma**2)), axis=(1, 2, 3))
        denom = y.shape[-1] * np.sum(np.log(sigma), axis=(1, 2, 3))
        return numer - denom

    def gauss_initialize_position(self,
                                  no_sample: np.ndarray,
                                  stddev: float = 0.02):
        """Initializes the position around a central value as mean sampled from a gauss distribution and updates self.x.
        Parameters:
        ----------
        no_sample: np.ndarray,
            Contains the number of samples to initialize under each mean. should be in the form [[3],[2]..], where here it means 3 samples and 2 samples around the first entry and second entry,respectively in self.central_value is taken.
        stddev: float, optional (default 0.02)
            contains the stddev with which the electrons' coordinates are initialized
        """

        mean = self.central_value[0]
        specific_sample = no_sample[0][0]
        ndim = np.shape(self.central_value)[1]
        self.x = np.random.normal(mean, stddev,
                                  (self.batch_no, specific_sample, 1, ndim))

        end = np.shape(self.central_value)[0]
        for i in range(1, end):
            mean = self.central_value[i]
            specific_sample = no_sample[i][0]
            self.x = np.append(
                self.x,
                np.random.normal(mean, stddev,
                                 (self.batch_no, specific_sample, 1, ndim)),
                axis=1)

    def move(self,
             stddev: float = 0.02,
             asymmetric_func: Optional[Callable[[np.ndarray],
                                                np.ndarray]] = None,
             index: Optional[int] = None) -> float:
        """Performs Metropolis-Hasting move for self.x(electrons). The type of moves to be followed -(simultaneous or single-electron, symmetric or asymmetric) have been specified when calling the class.
        The self.x array is replaced with a new array at the end of each step containing the new electron's positions.

        Parameters:
        -----------
        asymmetric_func: Callable[[np.ndarray],np.ndarray], optional(default None)
            Should be specified for an asymmetric move.The function should take in only 1 argument- y: a numpy array wrt to which mean should be calculated.
            This function should return the mean for the asymmetric proposal. For ferminet, this function is the harmonic mean of the distance between the electron and the nucleus.
        stddev: float, optional (default 0.02)
            Specifies the standard deviation in the case of symmetric moves and the scaling factor of the standard deviation matrix in the case of asymmetric moves.
        index: int, optional (default None)
            Specifies the index of the electron to be updated in the case of a single electron move.

        Return
        ------
        float
            accepted move ratio of the MCMC steps.
        """

        lp1 = self.f(self.x)  # log probability of self.x state

        if self.simultaneous:
            if self.symmetric:
                for i in range(self.steps):
                    x2 = np.random.normal(self.x, stddev, self.x.shape)
                    lp2 = self.f(x2)  # log probability of x2 state
                    ratio = lp2 - lp1
                    move_prob = np.log(
                        np.random.uniform(low=0,
                                          high=1.0,
                                          size=np.shape(self.x)[0]))
                    cond = move_prob < ratio
                    lp1 = np.where(cond, lp2, lp1)
                    self.x = np.where(cond[:, None, None, None], x2, self.x)
                    self.num_accept += np.sum(cond)

            elif asymmetric_func is not None:
                for i in range(self.steps):
                    std = stddev * asymmetric_func(self.x)
                    x2 = np.random.normal(self.x, std, self.x.shape)
                    lp2 = self.f(x2)  # log probability of x2 state
                    lq1 = self.log_prob_gaussian(self.x, x2,
                                                 std)  # forward probability
                    lq2 = self.log_prob_gaussian(
                        x2, self.x,
                        stddev * asymmetric_func(x2))  # backward probability
                    ratio = lp2 + lq2 - lq1 - lp1
                    move_prob = np.log(
                        np.random.uniform(low=0,
                                          high=1.0,
                                          size=np.shape(self.x)[0]))
                    cond = move_prob < ratio
                    self.x = np.where(cond[:, None, None, None], x2, self.x)
                    lp1 = np.where(cond, lp2, lp1)
                    self.num_accept += np.sum(cond)

        elif index is not None:
            index = int(index)
            x2 = np.copy(self.x)
            altered_shape = (self.batch_no, 1, np.shape(self.x)[3])

            if self.symmetric:
                for i in range(self.steps):
                    x2[:, index, :, :] = np.random.normal(x2[:, index, :, :],
                                                          stddev,
                                                          size=altered_shape)
                    lp2 = self.f(x2)  # log probability of x2 state
                    ratio = lp2 - lp1
                    move_prob = np.log(
                        np.random.uniform(low=0,
                                          high=1.0,
                                          size=np.shape(self.x)[0]))
                    cond = move_prob < ratio
                    lp1 = np.where(cond, lp2, lp1)
                    self.x = np.where(cond[:, None, None, None], x2, self.x)

            elif asymmetric_func is not None:
                init_dev = stddev * asymmetric_func(
                    self.x)  # initial standard deviation matrix
                for i in range(self.steps):
                    std = stddev * asymmetric_func(self.x[:, index, :, :])
                    x2[:, index, :, :] = np.random.normal(x2[:, index, :, :],
                                                          std,
                                                          size=altered_shape)
                    lp2 = self.f(x2)  # log probability of x2 state
                    init_dev[:, index, :, :] = std
                    lq1 = self.log_prob_gaussian(
                        self.x, x2, init_dev)  # forward probability
                    lq2 = self.log_prob_gaussian(
                        x2, self.x,
                        stddev * asymmetric_func(x2))  # backward probability
                    ratio = lp2 + lq2 - lq1 - lp1
                    move_prob = np.log(
                        np.random.uniform(low=0,
                                          high=1.0,
                                          size=np.shape(self.x)[0]))
                    cond = move_prob < ratio
                    self.x = np.where(cond[:, None, None, None], x2, self.x)
                    lp1 = np.where(cond, lp2, lp1)
                    self.num_accept += np.sum(cond)

        return self.num_accept / (
            (i + 1) * np.shape(self.x)[0])  # accepted move ratio



import tensorflow as tf
from collections.abc import Sequence as SequenceCollection

import logging
import deepchem as dc
from deepchem.metrics import to_one_hot
from deepchem.models import KerasModel
from deepchem.models.layers import Stack
from deepchem.models.losses import SoftmaxCrossEntropy, L2Loss
from typing import Tuple, Iterable, List

logger = logging.getLogger(__name__)


class RobustMultitaskClassifier(KerasModel):
    """Implements a neural network for robust multitasking.

    The key idea of this model is to have bypass layers that feed
    directly from features to task output. This might provide some
    flexibility toroute around challenges in multitasking with
    destructive interference.

    References
    ----------
    This technique was introduced in [1]_

    .. [1] Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 layer_sizes=[1000],
                 weight_init_stddevs=0.02,
                 bias_init_consts=1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 dropouts=0.5,
                 activation_fns=tf.nn.relu,
                 n_classes=2,
                 bypass_layer_sizes=[100],
                 bypass_weight_init_stddevs=[.02],
                 bypass_bias_init_consts=[1.],
                 bypass_dropouts=[.5],
                 **kwargs):
        """  Create a RobustMultitaskClassifier.

        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: int
            number of features
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes).  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        bias_init_consts: list or loat
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.
        n_classes: int
            the number of classes
        bypass_layer_sizes: list
            the size of each dense layer in the bypass network. The length of this list determines the number of bypass layers.
        bypass_weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of bypass layers.
            same requirements as weight_init_stddevs
        bypass_bias_init_consts: list or float
            the value to initialize the biases in bypass layers
            same requirements as bias_init_consts
        bypass_dropouts: list or float
            the dropout probablity to use for bypass layers.
            same requirements as dropouts
        """
        self.n_tasks = n_tasks
        self.n_features = n_features
        self.n_classes = n_classes
        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers
        if weight_decay_penalty != 0.0:
            if weight_decay_penalty_type == 'l1':
                regularizer = tf.keras.regularizers.l1(weight_decay_penalty)
            else:
                regularizer = tf.keras.regularizers.l2(weight_decay_penalty)
        else:
            regularizer = None

        n_bypass_layers = len(bypass_layer_sizes)
        if not isinstance(bypass_weight_init_stddevs, SequenceCollection):
            bypass_weight_init_stddevs = [bypass_weight_init_stddevs
                                         ] * n_bypass_layers
        if not isinstance(bypass_bias_init_consts, SequenceCollection):
            bypass_bias_init_consts = [bypass_bias_init_consts
                                      ] * n_bypass_layers
        if not isinstance(bypass_dropouts, SequenceCollection):
            bypass_dropouts = [bypass_dropouts] * n_bypass_layers
        bypass_activation_fns = [activation_fns[0]] * n_bypass_layers

        # Add the input features.
        mol_features = tf.keras.Input(shape=(n_features,))
        prev_layer = mol_features

        # Add the shared dense layers
        for size, weight_stddev, bias_const, dropout, activation_fn in zip(
                layer_sizes, weight_init_stddevs, bias_init_consts, dropouts,
                activation_fns):
            layer = tf.keras.layers.Dense(
                size,
                activation=activation_fn,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=weight_stddev),
                bias_initializer=tf.constant_initializer(value=bias_const),
                kernel_regularizer=regularizer)(prev_layer)
            if dropout > 0.0:
                layer = tf.keras.layers.Dropout(rate=dropout)(layer)
            prev_layer = layer
        top_multitask_layer = prev_layer

        task_outputs = []
        for i in range(self.n_tasks):
            prev_layer = mol_features
            # Add task-specific bypass layers
            for size, weight_stddev, bias_const, dropout, activation_fn in zip(
                    bypass_layer_sizes, bypass_weight_init_stddevs,
                    bypass_bias_init_consts, bypass_dropouts,
                    bypass_activation_fns):
                layer = tf.keras.layers.Dense(
                    size,
                    activation=activation_fn,
                    kernel_initializer=tf.keras.initializers.TruncatedNormal(
                        stddev=weight_stddev),
                    bias_initializer=tf.constant_initializer(value=bias_const),
                    kernel_regularizer=regularizer)(prev_layer)
                if dropout > 0.0:
                    layer = tf.keras.layers.Dropout(rate=dropout)(layer)
                prev_layer = layer
            top_bypass_layer = prev_layer

            if n_bypass_layers > 0:
                task_layer = tf.keras.layers.Concatenate(axis=1)(
                    [top_multitask_layer, top_bypass_layer])
            else:
                task_layer = top_multitask_layer

            task_out = tf.keras.layers.Dense(n_classes)(task_layer)
            task_outputs.append(task_out)

        logits = Stack(axis=1)(task_outputs)
        output = tf.keras.layers.Softmax()(logits)
        model = tf.keras.Model(inputs=mol_features, outputs=[output, logits])
        super(RobustMultitaskClassifier,
              self).__init__(model,
                             SoftmaxCrossEntropy(),
                             output_types=['prediction', 'loss'],
                             **kwargs)

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                if y_b is not None:
                    y_b = to_one_hot(y_b.flatten(), self.n_classes).reshape(
                        -1, self.n_tasks, self.n_classes)
                yield ([X_b], [y_b], [w_b])

    def create_estimator_inputs(self, feature_columns, weight_column, features,
                                labels, mode):
        tensors = {}
        for layer, column in zip(self.features, feature_columns):
            tensors[layer] = tf.feature_column.input_layer(features, [column])
        if weight_column is not None:
            tensors[self.task_weights[0]] = tf.feature_column.input_layer(
                features, [weight_column])
        if labels is not None:
            tensors[self.labels[0]] = tf.one_hot(tf.cast(labels, tf.int32),
                                                 self.n_classes)
        return tensors


class RobustMultitaskRegressor(KerasModel):
    """Implements a neural network for robust multitasking.

    The key idea of this model is to have bypass layers that feed
    directly from features to task output. This might provide some
    flexibility to route around challenges in multitasking with
    destructive interference.

    References
    ----------
    .. [1]   Ramsundar, Bharath, et al. "Is multitask deep learning practical for pharma?." Journal of chemical information and modeling 57.8 (2017): 2068-2076.

    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 layer_sizes=[1000],
                 weight_init_stddevs=0.02,
                 bias_init_consts=1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 dropouts=0.5,
                 activation_fns=tf.nn.relu,
                 bypass_layer_sizes=[100],
                 bypass_weight_init_stddevs=[.02],
                 bypass_bias_init_consts=[1.],
                 bypass_dropouts=[.5],
                 **kwargs):
        """ Create a RobustMultitaskRegressor.

        Parameters
        ----------
        n_tasks: int
            number of tasks
        n_features: int
            number of features
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes).  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        bias_init_consts: list or loat
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.
        bypass_layer_sizes: list
            the size of each dense layer in the bypass network. The length of this list determines the number of bypass layers.
        bypass_weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of bypass layers.
            same requirements as weight_init_stddevs
        bypass_bias_init_consts: list or float
            the value to initialize the biases in bypass layers
            same requirements as bias_init_consts
        bypass_dropouts: list or float
            the dropout probablity to use for bypass layers.
            same requirements as dropouts
        """
        self.n_tasks = n_tasks
        self.n_features = n_features
        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers
        if weight_decay_penalty != 0.0:
            if weight_decay_penalty_type == 'l1':
                regularizer = tf.keras.regularizers.l1(weight_decay_penalty)
            else:
                regularizer = tf.keras.regularizers.l2(weight_decay_penalty)
        else:
            regularizer = None

        n_bypass_layers = len(bypass_layer_sizes)
        if not isinstance(bypass_weight_init_stddevs, SequenceCollection):
            bypass_weight_init_stddevs = [bypass_weight_init_stddevs
                                         ] * n_bypass_layers
        if not isinstance(bypass_bias_init_consts, SequenceCollection):
            bypass_bias_init_consts = [bypass_bias_init_consts
                                      ] * n_bypass_layers
        if not isinstance(bypass_dropouts, SequenceCollection):
            bypass_dropouts = [bypass_dropouts] * n_bypass_layers
        bypass_activation_fns = [activation_fns[0]] * n_bypass_layers

        # Add the input features.
        mol_features = tf.keras.Input(shape=(n_features,))
        prev_layer = mol_features

        # Add the shared dense layers
        for size, weight_stddev, bias_const, dropout, activation_fn in zip(
                layer_sizes, weight_init_stddevs, bias_init_consts, dropouts,
                activation_fns):
            layer = tf.keras.layers.Dense(
                size,
                activation=activation_fn,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=weight_stddev),
                bias_initializer=tf.constant_initializer(value=bias_const),
                kernel_regularizer=regularizer)(prev_layer)
            if dropout > 0.0:
                layer = tf.keras.layers.Dropout(rate=dropout)(layer)
            prev_layer = layer
        top_multitask_layer = prev_layer

        task_outputs = []
        for i in range(self.n_tasks):
            prev_layer = mol_features
            # Add task-specific bypass layers
            for size, weight_stddev, bias_const, dropout, activation_fn in zip(
                    bypass_layer_sizes, bypass_weight_init_stddevs,
                    bypass_bias_init_consts, bypass_dropouts,
                    bypass_activation_fns):
                layer = tf.keras.layers.Dense(
                    size,
                    activation=activation_fn,
                    kernel_initializer=tf.keras.initializers.TruncatedNormal(
                        stddev=weight_stddev),
                    bias_initializer=tf.constant_initializer(value=bias_const),
                    kernel_regularizer=regularizer)(prev_layer)
                if dropout > 0.0:
                    layer = tf.keras.layers.Dropout(rate=dropout)(layer)
                prev_layer = layer
            top_bypass_layer = prev_layer

            if n_bypass_layers > 0:
                task_layer = tf.keras.layers.Concatenate(axis=1)(
                    [top_multitask_layer, top_bypass_layer])
            else:
                task_layer = top_multitask_layer

            task_out = tf.keras.layers.Dense(1)(task_layer)
            task_outputs.append(task_out)

        outputs = Stack(axis=1)(task_outputs)
        model = tf.keras.Model(inputs=mol_features, outputs=outputs)
        super(RobustMultitaskRegressor,
              self).__init__(model,
                             L2Loss(),
                             output_types=['prediction'],
                             **kwargs)

    def default_generator(
            self,
            dataset: dc.data.Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                yield ([X_b], [y_b], [w_b])



import unittest
import numpy as np
from deepchem.utils import geometry_utils
from deepchem.utils.geometry_utils import unit_vector
from deepchem.utils.geometry_utils import angle_between
from deepchem.utils.geometry_utils import compute_pairwise_distances
from deepchem.utils.geometry_utils import generate_random_unit_vector
from deepchem.utils.geometry_utils import generate_random_rotation_matrix
from deepchem.utils.geometry_utils import is_angle_within_cutoff


class TestGeometryUtils(unittest.TestCase):

    def test_generate_random_unit_vector(self):
        for _ in range(100):
            u = generate_random_unit_vector()
            # 3D vector with unit length
            self.assertEqual(u.shape, (3,))
            self.assertAlmostEqual(np.linalg.norm(u), 1.0)

    def test_generate_random_rotation_matrix(self):
        # very basic test, we check if rotations actually work in test_rotate_molecules
        for _ in range(100):
            m = generate_random_rotation_matrix()
            self.assertEqual(m.shape, (3, 3))

    def test_unit_vector(self):
        for _ in range(10):
            vector = np.random.rand(3)
            norm_vector = unit_vector(vector)
            self.assertAlmostEqual(np.linalg.norm(norm_vector), 1.0)

    def test_angle_between(self):
        for _ in range(10):
            v1 = np.random.rand(3,)
            v2 = np.random.rand(3,)
            angle = angle_between(v1, v2)
            self.assertLessEqual(angle, np.pi)
            self.assertGreaterEqual(angle, 0.0)
            self.assertAlmostEqual(angle_between(v1, v1), 0.0)
            self.assertAlmostEqual(angle_between(v1, -v1), np.pi)

    def test_is_angle_within_cutoff(self):
        v1 = np.array([1, 0, 0])
        v2 = np.array([-1, 0, 0])
        angle_cutoff = 10
        assert is_angle_within_cutoff(v1, v2, angle_cutoff)

    def test_compute_pairwise_distances(self):
        n1 = 10
        n2 = 50
        coords1 = np.random.rand(n1, 3)
        coords2 = np.random.rand(n2, 3)

        distance = compute_pairwise_distances(coords1, coords2)
        self.assertEqual(distance.shape, (n1, n2))
        self.assertTrue((distance >= 0).all())
        # random coords between 0 and 1, so the max possible distance in sqrt(3)
        self.assertTrue((distance <= 3.0**0.5).all())

        # check if correct distance metric was used
        coords1 = np.array([[0, 0, 0], [1, 0, 0]])
        coords2 = np.array([[1, 0, 0], [2, 0, 0], [3, 0, 0]])
        distance = compute_pairwise_distances(coords1, coords2)
        self.assertTrue((distance == [[1, 2, 3], [0, 1, 2]]).all())

    def test_compute_centroid(self):
        N = 10
        coords = np.random.rand(N, 3)
        centroid = geometry_utils.compute_centroid(coords)
        assert centroid.shape == (3,)

    def test_subract_centroid(self):
        N = 10
        coords = np.random.rand(N, 3)
        centroid = geometry_utils.compute_centroid(coords)
        new_coords = geometry_utils.subtract_centroid(coords, centroid)
        assert new_coords.shape == (N, 3)
        new_centroid = geometry_utils.compute_centroid(new_coords)
        assert new_centroid.shape == (3,)
        np.testing.assert_almost_equal(new_centroid,
                                       np.zeros_like(new_centroid))



"""
Utility functions to evaluate models on datasets.
"""
import csv
import logging
import warnings
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import deepchem as dc
from deepchem.metrics import Metric

logger = logging.getLogger(__name__)

Score = Dict[str, float]
Metric_Func = Callable[..., Any]
Metrics = Union[Metric, Metric_Func, List[Metric], List[Metric_Func]]


def output_statistics(scores: Score, stats_out: str) -> None:
    """Write computed stats to file.

    Statistics are written to specified `stats_out` file.

    Parameters
    ----------
    scores: dict
        Dictionary mapping names of metrics to scores.
    stats_out: str
        Name of file to write scores to.
    """
    logger.warning("output_statistics is deprecated.")
    with open(stats_out, "w") as statsfile:
        statsfile.write(str(scores) + "\n")


def output_predictions(dataset: "dc.data.Dataset", y_preds: np.ndarray,
                       csv_out: str) -> None:
    """Writes predictions to file.

    Writes predictions made on `dataset` to a specified file on
    disk. `dataset.ids` are used to format predictions. The produce CSV file will have format as follows

    | ID          | Task1Name    | Task2Name    |
    | ----------- | ------------ | ------------ |
    | identifer1  | prediction11 | prediction12 |
    | identifer2  | prediction21 | prediction22 |

    Parameters
    ----------
    dataset: dc.data.Dataset
        Dataset on which predictions have been made.
    y_preds: np.ndarray
        Predictions to output
    csv_out: str
        Name of file to write predictions to.
    """
    data_ids = dataset.ids
    n_tasks = len(dataset.get_task_names())
    y_preds = np.reshape(y_preds, (len(y_preds), n_tasks))
    assert len(y_preds) == len(data_ids)
    with open(csv_out, "w") as csvfile:
        csvwriter = csv.writer(csvfile)
        csvwriter.writerow(["ID"] + dataset.get_task_names())
        for mol_id, y_pred in zip(data_ids, y_preds):
            csvwriter.writerow([mol_id] + list(y_pred))


def _process_metric_input(metrics: Metrics) -> List[Metric]:
    """A private helper method which processes metrics correctly.

    Metrics can be input as `dc.metrics.Metric` objects, lists of
    `dc.metrics.Metric` objects, or as raw metric functions or lists of
    raw metric functions. Metric functions are functions which accept
    two arguments `y_true, y_pred` both of which must be `np.ndarray`
    objects and return a float value. This functions normalizes these
    different types of inputs to type `list[dc.metrics.Metric]` object
    for ease of later processing.

    Note that raw metric functions which don't have names attached will
    simply be named "metric-#" where # is their position in the provided
    metric list. For example, "metric-1" or "metric-7"

    Parameters
    ----------
    metrics: dc.metrics.Metric/list[dc.metrics.Metric]/metric function/ list[metric function]
        Input metrics to process.

    Returns
    -------
    final_metrics: list[dc.metrics.Metric]
        Converts all input metrics and outputs a list of
        `dc.metrics.Metric` objects.
    """
    # Make sure input is a list
    if not isinstance(metrics, list):
        # FIXME: Incompatible types in assignment
        metrics = [metrics]  # type: ignore

    final_metrics = []
    # FIXME: Argument 1 to "enumerate" has incompatible type
    for i, metric in enumerate(metrics):  # type: ignore
        # Ensure that metric is wrapped in a list.
        if isinstance(metric, Metric):
            final_metrics.append(metric)
        # This case checks if input is a function then wraps a
        # dc.metrics.Metric object around it
        elif callable(metric):
            wrap_metric = Metric(metric, name="metric-%d" % (i + 1))
            final_metrics.append(wrap_metric)
        else:
            raise ValueError(
                "metrics must be one of metric function / dc.metrics.Metric object /"
                "list of dc.metrics.Metric or metric functions.")
    return final_metrics


def relative_difference(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Compute the relative difference between x and y

    The two argument arrays must have the same shape.

    Parameters
    ----------
    x: np.ndarray
        First input array
    y: np.ndarray
        Second input array

    Returns
    -------
    z: np.ndarray
        We will have `z == np.abs(x-y) / np.abs(max(x, y))`.
    """
    warnings.warn(
        "Directly use `(x - y) / np.abs(y)` or `np.isclose`, `np.allclose` for testing tolerance",
        FutureWarning)
    z = (x - y) / abs(y)
    return z


class Evaluator(object):
    """Class that evaluates a model on a given dataset.

    The evaluator class is used to evaluate a `dc.models.Model` class on
    a given `dc.data.Dataset` object. The evaluator is aware of
    `dc.trans.Transformer` objects so will automatically undo any
    transformations which have been applied.

    Examples
    --------
    Evaluators allow for a model to be evaluated directly on a Metric
    for `sklearn`. Let's do a bit of setup constructing our dataset and
    model.

    >>> import deepchem as dc
    >>> import numpy as np
    >>> X = np.random.rand(10, 5)
    >>> y = np.random.rand(10, 1)
    >>> dataset = dc.data.NumpyDataset(X, y)
    >>> model = dc.models.MultitaskRegressor(1, 5)
    >>> transformers = []

    Then you can evaluate this model as follows
    >>> import sklearn
    >>> evaluator = Evaluator(model, dataset, transformers)
    >>> multitask_scores = evaluator.compute_model_performance(
    ...     sklearn.metrics.mean_absolute_error)

    Evaluators can also be used with `dc.metrics.Metric` objects as well
    in case you want to customize your metric further.

    >>> evaluator = Evaluator(model, dataset, transformers)
    >>> metric = dc.metrics.Metric(dc.metrics.mae_score)
    >>> multitask_scores = evaluator.compute_model_performance(metric)
    """

    def __init__(self, model, dataset: "dc.data.Dataset",
                 transformers: List["dc.trans.Transformer"]):
        """Initialize this evaluator

        Parameters
        ----------
        model: Model
            Model to evaluate. Note that this must be a regression or
            classification model and not a generative model.
        dataset: Dataset
            Dataset object to evaluate `model` on.
        transformers: List[Transformer]
            List of `dc.trans.Transformer` objects. These transformations
            must have been applied to `dataset` previously. The dataset will
            be untransformed for metric evaluation.
        """

        self.model = model
        self.dataset = dataset
        self.output_transformers = [
            transformer for transformer in transformers
            if transformer.transform_y
        ]

    def output_statistics(self, scores: Score, stats_out: str):
        """ Write computed stats to file.

        Parameters
        ----------
        scores: dict
            Dictionary mapping names of metrics to scores.
        stats_out: str
            Name of file to write scores to.
        """
        logger.warning(
            "Evaluator.output_statistics is deprecated."
            "Please use dc.utils.evaluate.output_statistics instead."
            "This method will be removed in a future version of DeepChem.")
        with open(stats_out, "w") as statsfile:
            statsfile.write(str(scores) + "\n")

    def output_predictions(self, y_preds: np.ndarray, csv_out: str):
        """Writes predictions to file.

        Writes predictions made on `self.dataset` to a specified file on
        disk. `self.dataset.ids` are used to format predictions.

        Parameters
        ----------
        y_preds: np.ndarray
            Predictions to output
        csv_out: str
            Name of file to write predictions to.
        """
        logger.warning(
            "Evaluator.output_predictions is deprecated."
            "Please use dc.utils.evaluate.output_predictions instead."
            "This method will be removed in a future version of DeepChem.")
        data_ids = self.dataset.ids
        n_tasks = len(self.dataset.get_task_names())
        y_preds = np.reshape(y_preds, (len(y_preds), n_tasks))
        assert len(y_preds) == len(data_ids)
        with open(csv_out, "w") as csvfile:
            csvwriter = csv.writer(csvfile)
            csvwriter.writerow(["ID"] + self.dataset.get_task_names())
            for mol_id, y_pred in zip(data_ids, y_preds):
                csvwriter.writerow([mol_id] + list(y_pred))

    def compute_model_performance(
            self,
            metrics: Metrics,
            csv_out: Optional[str] = None,
            stats_out: Optional[str] = None,
            per_task_metrics: bool = False,
            use_sample_weights: bool = False,
            n_classes: int = 2) -> Union[Score, Tuple[Score, Score]]:
        """
        Computes statistics of model on test data and saves results to csv.

        Parameters
        ----------
        metrics: dc.metrics.Metric/list[dc.metrics.Metric]/function
            The set of metrics provided. This class attempts to do some
            intelligent handling of input. If a single `dc.metrics.Metric`
            object is provided or a list is provided, it will evaluate
            `self.model` on these metrics. If a function is provided, it is
            assumed to be a metric function that this method will attempt to
            wrap in a `dc.metrics.Metric` object. A metric function must
            accept two arguments, `y_true, y_pred` both of which are
            `np.ndarray` objects and return a floating point score. The
            metric function may also accept a keyword argument
            `sample_weight` to account for per-sample weights.
        csv_out: str, optional (DEPRECATED)
            Filename to write CSV of model predictions.
        stats_out: str, optional (DEPRECATED)
            Filename to write computed statistics.
        per_task_metrics: bool, optional
            If true, return computed metric for each task on multitask dataset.
        use_sample_weights: bool, optional (default False)
            If set, use per-sample weights `w`.
        n_classes: int, optional (default None)
            If specified, will use `n_classes` as the number of unique classes
            in `self.dataset`. Note that this argument will be ignored for
            regression metrics.

        Returns
        -------
        multitask_scores: dict
            Dictionary mapping names of metrics to metric scores.
        all_task_scores: dict, optional
            If `per_task_metrics == True`, then returns a second dictionary
            of scores for each task separately.
        """
        if csv_out is not None:
            logger.warning(
                "csv_out is deprecated as an argument and will be removed in a future version of DeepChem."
                "Output is not written to CSV; manually write output instead.")
        if stats_out is not None:
            logger.warning(
                "stats_out is deprecated as an argument and will be removed in a future version of DeepChem."
                "Stats output is not written; please manually write output instead"
            )
        # Process input metrics
        metrics = _process_metric_input(metrics)

        y = self.dataset.y
        y = dc.trans.undo_transforms(y, self.output_transformers)
        w = self.dataset.w

        y_pred = self.model.predict(self.dataset, self.output_transformers)
        n_tasks = len(self.dataset.get_task_names())

        multitask_scores = {}
        all_task_scores = {}

        # Compute multitask metrics
        for metric in metrics:
            results = metric.compute_metric(
                y,
                y_pred,
                w,
                per_task_metrics=per_task_metrics,
                n_tasks=n_tasks,
                n_classes=n_classes,
                use_sample_weights=use_sample_weights)
            if per_task_metrics:
                multitask_scores[metric.name], computed_metrics = results
                all_task_scores[metric.name] = computed_metrics
            else:
                multitask_scores[metric.name] = results

        if not per_task_metrics:
            return multitask_scores
        else:
            return multitask_scores, all_task_scores


class GeneratorEvaluator(object):
    """Evaluate models on a stream of data.

    This class is a partner class to `Evaluator`. Instead of operating
    over datasets this class operates over a generator which yields
    batches of data to feed into provided model.

    Examples
    --------
    >>> import deepchem as dc
    >>> import numpy as np
    >>> X = np.random.rand(10, 5)
    >>> y = np.random.rand(10, 1)
    >>> dataset = dc.data.NumpyDataset(X, y)
    >>> model = dc.models.MultitaskRegressor(1, 5)
    >>> generator = model.default_generator(dataset, pad_batches=False)
    >>> transformers = []

    Then you can evaluate this model as follows

    >>> import sklearn
    >>> evaluator = GeneratorEvaluator(model, generator, transformers)
    >>> multitask_scores = evaluator.compute_model_performance(
    ...     sklearn.metrics.mean_absolute_error)

    Evaluators can also be used with `dc.metrics.Metric` objects as well
    in case you want to customize your metric further. (Note that a given
    generator can only be used once so we have to redefine the generator here.)

    >>> generator = model.default_generator(dataset, pad_batches=False)
    >>> evaluator = GeneratorEvaluator(model, generator, transformers)
    >>> metric = dc.metrics.Metric(dc.metrics.mae_score)
    >>> multitask_scores = evaluator.compute_model_performance(metric)
    """

    def __init__(self,
                 model,
                 generator: Iterable[Tuple[Any, Any, Any]],
                 transformers: List["dc.trans.Transformer"],
                 labels: Optional[List] = None,
                 weights: Optional[List] = None):
        """
        Parameters
        ----------
        model: Model
            Model to evaluate.
        generator: generator
            Generator which yields batches to feed into the model. For a
            KerasModel, it should be a tuple of the form (inputs, labels,
            weights). The "correct" way to create this generator is to use
            `model.default_generator` as shown in the example above.
        transformers: List[Transformer]
            Tranformers to "undo" when applied to the models outputs
        labels: list of Layer
            layers which are keys in the generator to compare to outputs
        weights: list of Layer
            layers which are keys in the generator for weight matrices
        """
        self.model = model
        self.generator = generator
        self.output_transformers = [
            transformer for transformer in transformers
            if transformer.transform_y
        ]
        self.label_keys = labels
        self.weights = weights
        if labels is not None and len(labels) != 1:
            raise ValueError(
                "GeneratorEvaluator currently only supports one label")

    def compute_model_performance(
            self,
            metrics: Metrics,
            per_task_metrics: bool = False,
            use_sample_weights: bool = False,
            n_classes: int = 2) -> Union[Score, Tuple[Score, Score]]:
        """
        Computes statistics of model on test data and saves results to csv.

        Parameters
        ----------
        metrics: dc.metrics.Metric/list[dc.metrics.Metric]/function
            The set of metrics provided. This class attempts to do some
            intelligent handling of input. If a single `dc.metrics.Metric`
            object is provided or a list is provided, it will evaluate
            `self.model` on these metrics. If a function is provided, it is
            assumed to be a metric function that this method will attempt to
            wrap in a `dc.metrics.Metric` object. A metric function must
            accept two arguments, `y_true, y_pred` both of which are
            `np.ndarray` objects and return a floating point score.
        per_task_metrics: bool, optional
            If true, return computed metric for each task on multitask
            dataset.
        use_sample_weights: bool, optional (default False)
            If set, use per-sample weights `w`.
        n_classes: int, optional (default None)
            If specified, will assume that all `metrics` are classification
            metrics and will use `n_classes` as the number of unique classes
            in `self.dataset`.

        Returns
        -------
        multitask_scores: dict
            Dictionary mapping names of metrics to metric scores.
        all_task_scores: dict, optional
            If `per_task_metrics == True`, then returns a second dictionary
            of scores for each task separately.
        """
        metrics = _process_metric_input(metrics)

        # We use y/w to aggregate labels/weights across generator.
        y = []
        w = []

        def generator_closure():
            """This function is used to pull true labels/weights out as we iterate over the generator."""
            if self.label_keys is None:
                weights = None
                # This is a KerasModel.
                for batch in self.generator:
                    # Some datasets have weights
                    try:
                        inputs, labels, weights = batch
                    except ValueError:
                        try:
                            inputs, labels, weights, ids = batch
                        except ValueError:
                            raise ValueError(
                                "Generator must yield values of form (input, labels, weights) or (input, labels, weights, ids)"
                            )
                    y.append(labels[0])
                    if len(weights) > 0:
                        w.append(weights[0])
                    yield (inputs, labels, weights)

        # Process predictions and populate y/w lists
        y_pred = self.model.predict_on_generator(generator_closure())

        # Combine labels/weights
        y = np.concatenate(y, axis=0)
        w = np.concatenate(w, axis=0)

        multitask_scores = {}
        all_task_scores = {}

        # Undo data transformations.
        y_true = dc.trans.undo_transforms(y, self.output_transformers)
        y_pred = dc.trans.undo_transforms(y_pred, self.output_transformers)

        # Compute multitask metrics
        for metric in metrics:
            results = metric.compute_metric(
                y_true,
                y_pred,
                w,
                per_task_metrics=per_task_metrics,
                n_classes=n_classes,
                use_sample_weights=use_sample_weights)
            if per_task_metrics:
                multitask_scores[metric.name], computed_metrics = results
                all_task_scores[metric.name] = computed_metrics
            else:
                multitask_scores[metric.name] = results

        if not per_task_metrics:
            return multitask_scores
        else:
            return multitask_scores, all_task_scores



"""
Process an input dataset into a format suitable for machine learning.
"""
import os
import tempfile
import zipfile
import time
import logging
import warnings
from typing import List, Optional, Tuple, Any, Sequence, Union, Iterator

import pandas as pd
import numpy as np
from deepchem.utils.typing import OneOrMany
from deepchem.utils.data_utils import load_image_files, load_csv_files, load_json_files, load_sdf_files, unzip_file
from deepchem.feat import UserDefinedFeaturizer, Featurizer
from deepchem.data import Dataset, DiskDataset, NumpyDataset, ImageDataset
from deepchem.feat.molecule_featurizers import OneHotFeaturizer
from deepchem.utils.genomics_utils import encode_bio_sequence

try:
    from deepchem.feat.dft_data import DFTEntry
    import yaml
    from yaml.loader import SafeLoader
except ModuleNotFoundError:
    pass

logger = logging.getLogger(__name__)


def _convert_df_to_numpy(df: pd.DataFrame,
                         tasks: List[str]) -> Tuple[np.ndarray, np.ndarray]:
    """Transforms a dataframe containing deepchem input into numpy arrays

    This is a private helper method intended to help parse labels and
    weights arrays from a pandas dataframe. Here `df` is a dataframe
    which has columns for each task in `tasks`. These labels are
    extracted into a labels array `y`. Weights `w` are initialized to
    all ones, but weights for any missing labels are set to 0.

    Parameters
    ----------
    df: pd.DataFrame
        Pandas dataframe with columns for all tasks
    tasks: List[str]
        List of tasks

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The tuple is `(w, y)`.
    """
    n_samples = df.shape[0]
    n_tasks = len(tasks)

    y = np.hstack([
        np.reshape(np.array(df[task].values), (n_samples, 1)) for task in tasks
    ])
    w = np.ones((n_samples, n_tasks))
    if y.dtype.kind in ['O', 'U']:
        missing = (y == '')
        y[missing] = 0
        w[missing] = 0

    return y.astype(float), w.astype(float)


class DataLoader(object):
    """Handles loading/featurizing of data from disk.

    The main use of `DataLoader` and its child classes is to make it
    easier to load large datasets into `Dataset` objects.`

    `DataLoader` is an abstract superclass that provides a
    general framework for loading data into DeepChem. This class should
    never be instantiated directly.  To load your own type of data, make
    a subclass of `DataLoader` and provide your own implementation for
    the `create_dataset()` method.

    To construct a `Dataset` from input data, first instantiate a
    concrete data loader (that is, an object which is an instance of a
    subclass of `DataLoader`) with a given `Featurizer` object. Then
    call the data loader's `create_dataset()` method on a list of input
    files that hold the source data to process. Note that each subclass
    of `DataLoader` is specialized to handle one type of input data so
    you will have to pick the loader class suitable for your input data
    type.

    Note that it isn't necessary to use a data loader to process input
    data. You can directly use `Featurizer` objects to featurize
    provided input into numpy arrays, but note that this calculation
    will be performed in memory, so you will have to write generators
    that walk the source files and write featurized data to disk
    yourself. `DataLoader` and its subclasses make this process easier
    for you by performing this work under the hood.
    """

    def __init__(self,
                 tasks: List[str],
                 featurizer: Featurizer,
                 id_field: Optional[str] = None,
                 log_every_n: int = 1000):
        """Construct a DataLoader object.

        This constructor is provided as a template mainly. You
        shouldn't ever call this constructor directly as a user.

        Parameters
        ----------
        tasks: List[str]
            List of task names
        featurizer: Featurizer
            Featurizer to use to process data.
        id_field: str, optional (default None)
            Name of field that holds sample identifier. Note that the
            meaning of "field" depends on the input data type and can have a
            different meaning in different subclasses. For example, a CSV
            file could have a field as a column, and an SDF file could have
            a field as molecular property.
        log_every_n: int, optional (default 1000)
            Writes a logging statement this often.
        """
        if self.__class__ is DataLoader:
            raise ValueError(
                "DataLoader should never be instantiated directly. Use a subclass instead."
            )
        if not isinstance(tasks, list):
            raise ValueError("tasks must be a list.")
        self.tasks = tasks
        self.id_field = id_field
        self.user_specified_features = None
        if isinstance(featurizer, UserDefinedFeaturizer):
            self.user_specified_features = featurizer.feature_fields
        self.featurizer = featurizer
        self.log_every_n = log_every_n

    def featurize(self,
                  inputs: OneOrMany[Any],
                  data_dir: Optional[str] = None,
                  shard_size: Optional[int] = 8192) -> Dataset:
        """Featurize provided files and write to specified location.

        DEPRECATED: This method is now a wrapper for `create_dataset()`
        and calls that method under the hood.

        For large datasets, automatically shards into smaller chunks
        for convenience. This implementation assumes that the helper
        methods `_get_shards` and `_featurize_shard` are implemented and
        that each shard returned by `_get_shards` is a pandas dataframe.
        You may choose to reuse or override this method in your subclass
        implementations.

        Parameters
        ----------
        inputs: List
            List of inputs to process. Entries can be filenames or arbitrary objects.
        data_dir: str, default None
            Directory to store featurized dataset.
        shard_size: int, optional (default 8192)
            Number of examples stored in each shard.

        Returns
        -------
        Dataset
            A `Dataset` object containing a featurized representation of data
            from `inputs`.
        """
        warnings.warn(
            "featurize() is deprecated and has been renamed to create_dataset()."
            "featurize() will be removed in DeepChem 3.0", FutureWarning)
        return self.create_dataset(inputs, data_dir, shard_size)

    def create_dataset(self,
                       inputs: OneOrMany[Any],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 8192) -> Dataset:
        """Creates and returns a `Dataset` object by featurizing provided files.

        Reads in `inputs` and uses `self.featurizer` to featurize the
        data in these inputs.  For large files, automatically shards
        into smaller chunks of `shard_size` datapoints for convenience.
        Returns a `Dataset` object that contains the featurized dataset.

        This implementation assumes that the helper methods `_get_shards`
        and `_featurize_shard` are implemented and that each shard
        returned by `_get_shards` is a pandas dataframe.  You may choose
        to reuse or override this method in your subclass implementations.

        Parameters
        ----------
        inputs: List
            List of inputs to process. Entries can be filenames or arbitrary objects.
        data_dir: str, optional (default None)
            Directory to store featurized dataset.
        shard_size: int, optional (default 8192)
            Number of examples stored in each shard.

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation of data
            from `inputs`.
        """
        logger.info("Loading raw samples now.")
        logger.info("shard_size: %s" % str(shard_size))

        # Special case handling of single input
        if not isinstance(inputs, list):
            inputs = [inputs]

        def shard_generator():
            for shard_num, shard in enumerate(
                    self._get_shards(inputs, shard_size)):
                time1 = time.time()
                X, valid_inds = self._featurize_shard(shard)
                ids = shard[self.id_field].values
                ids = ids[valid_inds]
                if len(self.tasks) > 0:
                    # Featurize task results iff they exist.
                    y, w = _convert_df_to_numpy(shard, self.tasks)
                    # Filter out examples where featurization failed.
                    y, w = (y[valid_inds], w[valid_inds])
                    assert len(X) == len(ids) == len(y) == len(w)
                else:
                    # For prospective data where results are unknown, it
                    # makes no sense to have y values or weights.
                    y, w = (None, None)
                    assert len(X) == len(ids)

                time2 = time.time()
                logger.info("TIMING: featurizing shard %d took %0.3f s" %
                            (shard_num, time2 - time1))
                yield X, y, w, ids

        return DiskDataset.create_dataset(shard_generator(), data_dir,
                                          self.tasks)

    def _get_shards(self, inputs: List, shard_size: Optional[int]) -> Iterator:
        """Stub for children classes.

        Should implement a generator that walks over the source data in
        `inputs` and returns a "shard" at a time. Here a shard is a
        chunk of input data that can reasonably be handled in memory. For
        example, this may be a set of rows from a CSV file or a set of
        molecules from a SDF file. To re-use the
        `DataLoader.create_dataset()` method, each shard must be a pandas
        dataframe.

        If you chose to override `create_dataset()` directly you don't
        need to override this helper method.

        Parameters
        ----------
        inputs: list
            List of inputs to process. Entries can be filenames or arbitrary objects.
        shard_size: int, optional
            Number of examples stored in each shard.
        """
        raise NotImplementedError

    def _featurize_shard(self, shard: Any):
        """Featurizes a shard of input data.

        Recall a shard is a chunk of input data that can reasonably be
        handled in memory. For example, this may be a set of rows from a
        CSV file or a set of molecules from a SDF file. Featurize this
        shard in memory and return the results.

        Parameters
        ----------
        shard: Any
            A chunk of input data
        """
        raise NotImplementedError


class CSVLoader(DataLoader):
    """
    Creates `Dataset` objects from input CSV files.

    This class provides conveniences to load data from CSV files.
    It's possible to directly featurize data from CSV files using
    pandas, but this class may prove useful if you're processing
    large CSV files that you don't want to manipulate directly in
    memory.

    Examples
    --------
    Let's suppose we have some smiles and labels

    >>> smiles = ["C", "CCC"]
    >>> labels = [1.5, 2.3]

    Let's put these in a dataframe.

    >>> import pandas as pd
    >>> df = pd.DataFrame(list(zip(smiles, labels)), columns=["smiles", "task1"])

    Let's now write this to disk somewhere. We can now use `CSVLoader` to
    process this CSV dataset.

    >>> import tempfile
    >>> import deepchem as dc
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.CSVLoader(["task1"], feature_field="smiles",
    ...                              featurizer=dc.feat.CircularFingerprint())
    ...     dataset = loader.create_dataset(tmpfile.name)
    >>> len(dataset)
    2

    Of course in practice you should already have your data in a CSV file if
    you're using `CSVLoader`. If your data is already in memory, use
    `InMemoryLoader` instead.

    Sometimes there will be datasets without specific tasks, for example
    datasets which are used in unsupervised learning tasks. Such datasets
    can be loaded by leaving the `tasks` field empty.

    Example
    -------

    >>> x1, x2 = [2, 3, 4], [4, 6, 8]
    >>> df = pd.DataFrame({"x1":x1, "x2": x2}).reset_index()
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.CSVLoader(tasks=[], id_field="index", feature_field=["x1", "x2"],
    ...                              featurizer=dc.feat.DummyFeaturizer())
    ...     dataset = loader.create_dataset(tmpfile.name)
    >>> len(dataset)
    3

    """

    def __init__(self,
                 tasks: List[str],
                 featurizer: Featurizer,
                 feature_field: Optional[str] = None,
                 id_field: Optional[str] = None,
                 smiles_field: Optional[str] = None,
                 log_every_n: int = 1000):
        """Initializes CSVLoader.

        Parameters
        ----------
        tasks: List[str]
            List of task names
        featurizer: Featurizer
            Featurizer to use to process data.
        feature_field: str, optional (default None)
            Field with data to be featurized.
        id_field: str, optional, (default None)
            CSV column that holds sample identifier
        smiles_field: str, optional (default None) (DEPRECATED)
            Name of field that holds smiles string.
        log_every_n: int, optional (default 1000)
            Writes a logging statement this often.
        """
        if not isinstance(tasks, list):
            raise ValueError("tasks must be a list.")
        if smiles_field is not None:
            logger.warning(
                "smiles_field is deprecated and will be removed in a future version of DeepChem."
                "Use feature_field instead.")
            if feature_field is not None and smiles_field != feature_field:
                raise ValueError(
                    "smiles_field and feature_field if both set must have the same value."
                )
            elif feature_field is None:
                feature_field = smiles_field

        self.tasks = tasks
        self.feature_field = feature_field
        self.id_field = id_field
        if id_field is None:
            self.id_field = feature_field  # Use features as unique ids if necessary
        else:
            self.id_field = id_field
        self.user_specified_features = None
        if isinstance(featurizer, UserDefinedFeaturizer):
            self.user_specified_features = featurizer.feature_fields
        self.featurizer = featurizer
        self.log_every_n = log_every_n

    def _get_shards(self, input_files: List[str],
                    shard_size: Optional[int]) -> Iterator[pd.DataFrame]:
        """Defines a generator which returns data for each shard

        Parameters
        ----------
        input_files: List[str]
            List of filenames to process
        shard_size: int, optional
            The size of a shard of data to process at a time.

        Returns
        -------
        Iterator[pd.DataFrame]
            Iterator over shards
        """
        return load_csv_files(input_files, shard_size)

    def _featurize_shard(self,
                         shard: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Featurizes a shard of an input dataframe.

        Parameters
        ----------
        shard: pd.DataFrame
            DataFrame that holds a shard of the input CSV file

        Returns
        -------
        features: np.ndarray
            Features computed from CSV file.
        valid_inds: np.ndarray
            Indices of rows in source CSV with valid data.
        """
        logger.info("About to featurize shard.")
        if self.featurizer is None:
            raise ValueError(
                "featurizer must be specified in constructor to featurizer data/"
            )
        features = [elt for elt in self.featurizer(shard[self.feature_field])]
        valid_inds = np.array(
            [1 if np.array(elt).size > 0 else 0 for elt in features],
            dtype=bool)
        features = [
            elt for (is_valid, elt) in zip(valid_inds, features) if is_valid
        ]
        return np.array(features), valid_inds


class UserCSVLoader(CSVLoader):
    """
    Handles loading of CSV files with user-defined features.

    This is a convenience class that allows for descriptors already present in a
    CSV file to be extracted without any featurization necessary.

    Examples
    --------
    Let's suppose we have some descriptors and labels. (Imagine that these
    descriptors have been computed by an external program.)

    >>> desc1 = [1, 43]
    >>> desc2 = [-2, -22]
    >>> labels = [1.5, 2.3]
    >>> ids = ["cp1", "cp2"]

    Let's put these in a dataframe.

    >>> import pandas as pd
    >>> df = pd.DataFrame(list(zip(ids, desc1, desc2, labels)), columns=["id", "desc1", "desc2", "task1"])

    Let's now write this to disk somewhere. We can now use `UserCSVLoader` to
    process this CSV dataset.

    >>> import tempfile
    >>> import deepchem as dc
    >>> featurizer = dc.feat.UserDefinedFeaturizer(["desc1", "desc2"])
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.UserCSVLoader(["task1"], id_field="id",
    ...                              featurizer=featurizer)
    ...     dataset = loader.create_dataset(tmpfile.name)
    >>> len(dataset)
    2
    >>> dataset.X[0, 0]
    1

    The difference between `UserCSVLoader` and `CSVLoader` is that our
    descriptors (our features) have already been computed for us, but are spread
    across multiple columns of the CSV file.

    Of course in practice you should already have your data in a CSV file if
    you're using `UserCSVLoader`. If your data is already in memory, use
    `InMemoryLoader` instead.
    """

    def _get_shards(self, input_files: List[str],
                    shard_size: Optional[int]) -> Iterator[pd.DataFrame]:
        """Defines a generator which returns data for each shard

        Parameters
        ----------
        input_files: List[str]
            List of filenames to process
        shard_size: int, optional
            The size of a shard of data to process at a time.

        Returns
        -------
        Iterator[pd.DataFrame]
            Iterator over shards
        """
        return load_csv_files(input_files, shard_size)

    def _featurize_shard(self,
                         shard: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Featurizes a shard of an input dataframe.

        Parameters
        ----------
        shard: pd.DataFrame
            DataFrame that holds a shard of the input CSV file

        Returns
        -------
        features: np.ndarray
            Features extracted from CSV file.
        valid_inds: np.ndarray
            Indices of rows in source CSV with valid data.
        """
        assert isinstance(self.featurizer, UserDefinedFeaturizer)
        time1 = time.time()
        feature_fields = self.featurizer.feature_fields
        shard[feature_fields] = shard[feature_fields].apply(pd.to_numeric)
        X_shard = shard[feature_fields].to_numpy()
        time2 = time.time()
        logger.info("TIMING: user specified processing took %0.3f s" %
                    (time2 - time1))
        return (X_shard, np.ones(len(X_shard), dtype=bool))


class JsonLoader(DataLoader):
    """
    Creates `Dataset` objects from input json files.

    This class provides conveniences to load data from json files.
    It's possible to directly featurize data from json files using
    pandas, but this class may prove useful if you're processing
    large json files that you don't want to manipulate directly in
    memory.

    It is meant to load JSON files formatted as "records" in line
    delimited format, which allows for sharding.
    ``list like [{column -> value}, ... , {column -> value}]``.

    Examples
    --------
    Let's create the sample dataframe.

    >>> composition = ["LiCoO2", "MnO2"]
    >>> labels = [1.5, 2.3]
    >>> import pandas as pd
    >>> df = pd.DataFrame(list(zip(composition, labels)), columns=["composition", "task"])

    Dump the dataframe to the JSON file formatted as "records" in line delimited format and
    load the json file by JsonLoader.

    >>> import tempfile
    >>> import deepchem as dc
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_json(tmpfile.name, orient='records', lines=True)
    ...     featurizer = dc.feat.ElementPropertyFingerprint()
    ...     loader = dc.data.JsonLoader(["task"], feature_field="composition", featurizer=featurizer)
    ...     dataset = loader.create_dataset(tmpfile.name)
    >>> len(dataset)
    2
    """

    def __init__(self,
                 tasks: List[str],
                 feature_field: str,
                 featurizer: Featurizer,
                 label_field: Optional[str] = None,
                 weight_field: Optional[str] = None,
                 id_field: Optional[str] = None,
                 log_every_n: int = 1000):
        """Initializes JsonLoader.

        Parameters
        ----------
        tasks: List[str]
            List of task names
        feature_field: str
            JSON field with data to be featurized.
        featurizer: Featurizer
            Featurizer to use to process data
        label_field: str, optional (default None)
            Field with target variables.
        weight_field: str, optional (default None)
            Field with weights.
        id_field: str, optional (default None)
            Field for identifying samples.
        log_every_n: int, optional (default 1000)
            Writes a logging statement this often.
        """
        if not isinstance(tasks, list):
            raise ValueError("Tasks must be a list.")
        self.tasks = tasks
        self.feature_field = feature_field
        self.label_field = label_field
        self.weight_field = weight_field
        self.id_field = id_field

        self.user_specified_features = None
        if isinstance(featurizer, UserDefinedFeaturizer):
            self.user_specified_features = featurizer.feature_fields
        self.featurizer = featurizer
        self.log_every_n = log_every_n

    def create_dataset(self,
                       input_files: OneOrMany[str],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 8192) -> DiskDataset:
        """Creates a `Dataset` from input JSON files.

        Parameters
        ----------
        input_files: OneOrMany[str]
            List of JSON filenames.
        data_dir: Optional[str], default None
            Name of directory where featurized data is stored.
        shard_size: int, optional (default 8192)
            Shard size when loading data.

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation of data
            from `input_files`.
        """
        if not isinstance(input_files, list):
            try:
                if isinstance(input_files, str):
                    input_files = [input_files]
                else:
                    input_files = list(input_files)
            except TypeError:
                raise ValueError(
                    "input_files is of an unrecognized form. Must be one filename or a list of filenames."
                )

        def shard_generator():
            """Yield X, y, w, and ids for shards."""
            for shard_num, shard in enumerate(
                    self._get_shards(input_files, shard_size)):

                time1 = time.time()
                X, valid_inds = self._featurize_shard(shard)
                if self.id_field:
                    ids = shard[self.id_field].values
                else:
                    ids = np.ones(len(valid_inds))
                ids = ids[valid_inds]

                if len(self.tasks) > 0:
                    # Featurize task results if they exist.
                    y, w = _convert_df_to_numpy(shard, self.tasks)

                    if self.label_field:
                        y = shard[self.label_field]
                    if self.weight_field:
                        w = shard[self.weight_field]

                    # Filter out examples where featurization failed.
                    y, w = (y[valid_inds], w[valid_inds])
                    assert len(X) == len(ids) == len(y) == len(w)
                else:
                    # For prospective data where results are unknown, it
                    # makes no sense to have y values or weights.
                    y, w = (None, None)
                    assert len(X) == len(ids)

                time2 = time.time()
                logger.info("TIMING: featurizing shard %d took %0.3f s" %
                            (shard_num, time2 - time1))
                yield X, y, w, ids

        return DiskDataset.create_dataset(shard_generator(), data_dir,
                                          self.tasks)

    def _get_shards(self, input_files: List[str],
                    shard_size: Optional[int]) -> Iterator[pd.DataFrame]:
        """Defines a generator which returns data for each shard

        Parameters
        ----------
        input_files: List[str]
            List of filenames to process
        shard_size: int, optional
            The size of a shard of data to process at a time.

        Returns
        -------
        Iterator[pd.DataFrame]
            Iterator over shards
        """
        return load_json_files(input_files, shard_size)

    def _featurize_shard(self,
                         shard: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Featurizes a shard of an input dataframe.

        Helper that computes features for the given shard of data.

        Parameters
        ----------
        shard: pd.DataFrame
            DataFrame that holds data to be featurized.

        Returns
        -------
        features: np.ndarray
            Array of feature vectors. Note that samples for which featurization has
            failed will be filtered out.
        valid_inds: np.ndarray
            Boolean values indicating successful featurization for corresponding
            sample in the source.
        """
        logger.info("About to featurize shard.")
        if self.featurizer is None:
            raise ValueError(
                "featurizer must be specified in constructor to featurizer data/"
            )
        features = [elt for elt in self.featurizer(shard[self.feature_field])]
        valid_inds = np.array(
            [1 if np.array(elt).size > 0 else 0 for elt in features],
            dtype=bool)
        features = [
            elt for (is_valid, elt) in zip(valid_inds, features) if is_valid
        ]
        return np.array(features), valid_inds


class SDFLoader(DataLoader):
    """Creates a `Dataset` object from SDF input files.

    This class provides conveniences to load and featurize data from
    Structure Data Files (SDFs). SDF is a standard format for structural
    information (3D coordinates of atoms and bonds) of molecular compounds.

    Examples
    --------
    >>> import deepchem as dc
    >>> import os
    >>> current_dir = os.path.dirname(os.path.realpath(__file__))
    >>> featurizer = dc.feat.CircularFingerprint(size=16)
    >>> loader = dc.data.SDFLoader(["LogP(RRCK)"], featurizer=featurizer, sanitize=True)
    >>> dataset = loader.create_dataset(os.path.join(current_dir, "tests", "membrane_permeability.sdf")) # doctest:+ELLIPSIS
    >>> len(dataset)
    2
    """

    def __init__(self,
                 tasks: List[str],
                 featurizer: Featurizer,
                 sanitize: bool = False,
                 log_every_n: int = 1000):
        """Initialize SDF Loader

        Parameters
        ----------
        tasks: list[str]
            List of tasknames. These will be loaded from the SDF file.
        featurizer: Featurizer
            Featurizer to use to process data
        sanitize: bool, optional (default False)
            Whether to sanitize molecules.
        log_every_n: int, optional (default 1000)
            Writes a logging statement this often.
        """
        self.featurizer = featurizer
        self.sanitize = sanitize
        self.tasks = tasks
        # The field in which dc.utils.save.load_sdf_files stores RDKit mol objects
        self.mol_field = "mol"
        # The field in which load_sdf_files return value stores smiles
        self.id_field = "smiles"
        self.log_every_n = log_every_n

    def create_dataset(self,
                       inputs: OneOrMany[Any],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 8192) -> Dataset:
        """Creates and returns a `Dataset` object by featurizing provided sdf files.

        Parameters
        ----------
        inputs: List
            List of inputs to process. Entries can be filenames or arbitrary objects.
            Each file should be supported format (.sdf) or compressed folder of
            .sdf files
        data_dir: str, optional (default None)
            Directory to store featurized dataset.
        shard_size: int, optional (default 8192)
            Number of examples stored in each shard.

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation of data
            from `inputs`.
        """
        logger.info("Loading raw samples now.")
        logger.info("shard_size: %s" % str(shard_size))

        # Special case handling of single input
        if not isinstance(inputs, list):
            inputs = [inputs]

        processed_files = []
        for input_file in inputs:
            filename, extension = os.path.splitext(input_file)
            extension = extension.lower()
            if extension == ".sdf":
                processed_files.append(input_file)
            elif extension == ".zip":
                zip_dir = tempfile.mkdtemp()
                unzip_file(input_file, zip_dir)
                zip_files = [
                    os.path.join(zip_dir, name) for name in os.listdir(zip_dir)
                ]
                for zip_file in zip_files:
                    _, extension = os.path.splitext(zip_file)
                    extension = extension.lower()
                    if extension in [".sdf"]:
                        processed_files.append(zip_file)
            else:
                raise ValueError("Unsupported file format")

        inputs = processed_files

        def shard_generator():
            for shard_num, shard in enumerate(
                    self._get_shards(inputs, shard_size)):
                time1 = time.time()
                X, valid_inds = self._featurize_shard(shard)
                ids = shard[self.id_field].values
                ids = ids[valid_inds]
                if len(self.tasks) > 0:
                    # Featurize task results iff they exist.
                    y, w = _convert_df_to_numpy(shard, self.tasks)
                    # Filter out examples where featurization failed.
                    y, w = (y[valid_inds], w[valid_inds])
                    assert len(X) == len(ids) == len(y) == len(w)
                else:
                    # For prospective data where results are unknown, it
                    # makes no sense to have y values or weights.
                    y, w = (None, None)
                    assert len(X) == len(ids)

                time2 = time.time()
                logger.info("TIMING: featurizing shard %d took %0.3f s" %
                            (shard_num, time2 - time1))
                yield X, y, w, ids

        return DiskDataset.create_dataset(shard_generator(), data_dir,
                                          self.tasks)

    def _get_shards(self, input_files: List[str],
                    shard_size: Optional[int]) -> Iterator[pd.DataFrame]:
        """Defines a generator which returns data for each shard

        Parameters
        ----------
        input_files: List[str]
            List of filenames to process
        shard_size: int, optional
            The size of a shard of data to process at a time.

        Returns
        -------
        Iterator[pd.DataFrame]
            Iterator over shards
        """
        return load_sdf_files(input_files=input_files,
                              clean_mols=self.sanitize,
                              tasks=self.tasks,
                              shard_size=shard_size)

    def _featurize_shard(self,
                         shard: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Featurizes a shard of an input dataframe.

        Helper that computes features for the given shard of data.

        Parameters
        ----------
        shard: pd.DataFrame
            DataFrame that holds data to be featurized.

        Returns
        -------
        features: np.ndarray
            Array of feature vectors. Note that samples for which featurization has
            failed will be filtered out.
        valid_inds: np.ndarray
            Boolean values indicating successful featurization for corresponding
            sample in the source.
        """
        pos_cols = ['pos_x', 'pos_y', 'pos_z']
        if set(pos_cols).issubset(shard.columns):
            features = [
                elt for elt in self.featurizer(shard[self.mol_field],
                                               pos_x=shard['pos_x'],
                                               pos_y=shard['pos_y'],
                                               pos_z=shard['pos_z'])
            ]
        else:
            features = [elt for elt in self.featurizer(shard[self.mol_field])]
        valid_inds = np.array(
            [1 if np.array(elt).size > 0 else 0 for elt in features],
            dtype=bool)
        features = [
            elt for (is_valid, elt) in zip(valid_inds, features) if is_valid
        ]
        return np.array(features), valid_inds


class FASTALoader(DataLoader):
    """Handles loading of FASTA files.

    FASTA files are commonly used to hold sequence data. This
    class provides convenience files to lead FASTA data and
    one-hot encode the genomic sequences for use in downstream
    learning tasks.
    """

    def __init__(self,
                 featurizer: Optional[Featurizer] = None,
                 auto_add_annotations: bool = False,
                 legacy: bool = True):
        """Initialize FASTALoader.

        Parameters
        ----------
        featurizer: Featurizer (default: None)
            The Featurizer to be used for the loaded FASTA data.

            If featurizer is None and legacy is True, the original featurization
            logic is used, creating a one hot encoding of all included FASTA strings
            of shape
            (number of FASTA sequences, number of channels + 1, sequence length, 1).

            If featurizer is None and legacy is False, the featurizer is initialized
            as a OneHotFeaturizer object with charset ("A", "C", "T", "G") and
            max_length = None.

        auto_add_annotations: bool (default False)
            Whether create_dataset will automatically add [CLS] and [SEP] annotations
            to the sequences it reads in order to assist tokenization.
            Keep False if your FASTA file already includes [CLS] and [SEP] annotations.

        legacy: bool (default True)
            Whether to use legacy logic for featurization. Legacy mode will create
            a one hot encoding of the FASTA content of shape
            (number of FASTA sequences, number of channels + 1, max length, 1).

            Legacy mode is only tested for ACTGN charsets, and will be deprecated.
       """

        # Process legacy toggle
        if legacy:
            warnings.warn(
                """
                    Legacy mode is deprecated and will be removed in
                    DeepChem 3.0. Disable legacy mode by passing legacy=False
                    during construction of FASTALoader object.
                    """, FutureWarning)
            if featurizer is not None or auto_add_annotations:
                raise ValueError(f"""
                          featurizer option must be None and
                          auto_add_annotations must be false when legacy mode
                          is enabled. You set featurizer to {featurizer} and
                          auto_add_annotations to {auto_add_annotations}.
                          """)

        # Set attributes
        self.legacy = legacy
        self.auto_add_annotations = auto_add_annotations

        self.user_specified_features = None

        # Handle special featurizer cases
        if isinstance(featurizer,
                      UserDefinedFeaturizer):  # User defined featurizer
            self.user_specified_features = featurizer.feature_fields
        elif featurizer is None:  # Default featurizer
            featurizer = OneHotFeaturizer(charset=["A", "C", "T", "G"],
                                          max_length=None)

        # Set self.featurizer
        self.featurizer = featurizer

    def create_dataset(self,
                       input_files: OneOrMany[str],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = None) -> DiskDataset:
        """Creates a `Dataset` from input FASTA files.

        At present, FASTA support is limited and doesn't allow for sharding.

        Parameters
        ----------
        input_files: List[str]
            List of fasta files.
        data_dir: str, optional (default None)
            Name of directory where featurized data is stored.
        shard_size: int, optional (default None)
            For now, this argument is ignored and each FASTA file gets its
            own shard.

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation of data
            from `input_files`.
        """
        if isinstance(input_files, str):
            input_files = [input_files]

        def shard_generator():  # TODO Enable sharding with shard size parameter
            for input_file in input_files:
                if self.legacy:
                    X = encode_bio_sequence(input_file)
                else:
                    sequences = _read_file(input_file)
                    X = self.featurizer(sequences)
                ids = np.ones(len(X))
                # (X, y, w, ids)
                yield X, None, None, ids

        def _read_file(input_file: str):
            """
            Convert the FASTA file to a numpy array of FASTA-format strings.
            """

            # TODO don't convert all sequences into np array (allow shards)
            def _generate_sequences(fasta_file, header_mark=">") -> np.ndarray:
                """
                Uses a fasta_file to create a numpy array of annotated FASTA-format strings
                """
                sequences: np.ndarray = np.array([])
                sequence: np.ndarray = np.array([])
                header_read = False
                for line in fasta_file:
                    # Check if line is a header
                    if line.startswith(header_mark):  # New header line
                        header_read = True
                        sequences = _add_sequence(sequences, sequence)
                        sequence = np.array([])
                    elif header_read:  # Line contains sequence in FASTA format
                        if line[-1:] == '\n':  # Check last character in string
                            line = line[0:-1]  # Remove last character
                        sequence = np.append(sequence, line)
                sequences = _add_sequence(sequences,
                                          sequence)  # Add last sequence
                return sequences

            def _add_sequence(sequences: np.ndarray,
                              sequence: np.ndarray) -> np.ndarray:
                # Handle empty sequence
                if sequence is None or len(sequence) <= 0:
                    return np.array([])
                # Annotate start/stop of sequence
                if self.auto_add_annotations:
                    sequence = np.insert(sequence, 0, "[CLS]")
                    sequence = np.append(sequence, "[SEP]")
                new_sequence = ''.join(sequence)
                new_sequences = np.append(sequences, new_sequence)
                return new_sequences

            with open(input_file, 'r') as f:  # Read FASTA file
                return _generate_sequences(f)

        return DiskDataset.create_dataset(shard_generator(), data_dir)


def _fastq_load_files(input_files: List[str],
                      shard_size: Optional[int] = 4096) -> Iterator:
    """Load data as Iterator.

    Parameters
    ----------
    input_files: List[str]
        List of fastq filenames.
    shard_size: int, optional (default 4096)
        Chunksize for reading fastq files.

    Yields
    -------
    Iterator
        Generator which yields the data which is the same shard size.
    """
    shard_num = 0
    for input_file in input_files:
        logger.info("About to start loading fastq from %s." % input_file)
        # Open index file
        with open(input_file, 'r') as f:
            # create an empty list to store lines in files.
            df = []
            line_number = 0
            # iterate through each line in the input file
            for num, line in enumerate(f):
                # If the number of lines iterated through is equal or less than the shard size:
                if (shard_size is not None) and ((num + 1) - line_number <=
                                                 (shard_size * 4)):
                    # append to list
                    df.append(line)
                else:
                    # else yield the list
                    shard_num += 1
                    logger.info("Loading shard %d of size %s." %
                                (shard_num, str(shard_size)))
                    # set the line_number variable to the last line number (num) before 'yield' was called
                    line_number = num
                    # yield list (shard/batch)
                    yield df
                    # Re-initialize list with the index line to begin a new shard.
                    df = [line]
            if len(df) > 0:
                yield df


class FASTQLoader(DataLoader):
    """Handles loading of FASTQ files.

    FASTQ files are commonly used to hold very large sequence data. It is a variant of FASTA format.
    This class provides convenience files to load FASTQ data and one-hot encode
    the genomic sequences for use in downstream learning tasks.

    Example
    -------
    >>> import os
    >>> from deepchem.feat.molecule_featurizers import OneHotFeaturizer
    >>> from deepchem.data.data_loader import FASTQLoader
    >>> current_dir = os.path.dirname(os.path.abspath(__file__))
    >>> input_file = os.path.join(current_dir, "tests", "sample1.fastq")
    >>> loader = FASTQLoader()
    >>> sequences = loader.create_dataset(input_file)

    See Also
    --------
    `Info on the structure of FASTQ files <https://support.illumina.com/bulletins/2016/04/fastq-files-explained.html>`
    """

    def __init__(self,
                 featurizer: Optional[Featurizer] = None,
                 auto_add_annotations: bool = False,
                 return_quality_scores: bool = False):
        """Initialize FASTQLoader.

        Parameters
        ----------
        featurizer: Featurizer (default: None)
            The Featurizer to be used for the loaded FASTQ data.
            The featurizer is initialized as a OneHotFeaturizer object with charset ("A", "C", "T", "G") and
            max_length = None.
        auto_add_annotations: bool (default False)
            Whether create_dataset will automatically add [CLS] and [SEP] annotations
            to the sequences it reads in order to assist tokenization.
            Keep False if your FASTQ file already includes [CLS] and [SEP] annotations.
        return_quality_scores: bool (default True)
            returns the quality (likelihood) score of the nucleotides in the sequence.
       """

        # Set attributes
        self.auto_add_annotations = auto_add_annotations

        self.user_specified_features = None

        # Handle special featurizer cases
        if isinstance(featurizer,
                      UserDefinedFeaturizer):  # User defined featurizer
            self.user_specified_features = featurizer.feature_fields
        elif featurizer is None:  # Default featurizer
            featurizer = OneHotFeaturizer(charset=["A", "C", "T", "G"],
                                          max_length=None)

        # Set self.featurizer
        self.featurizer = featurizer
        # Set self.return_quality_scores
        self.return_quality_scores = return_quality_scores

    def _get_shards(self,
                    input_files: List[str],
                    shard_size: Optional[int] = 4096) -> Iterator:
        """Defines a generator which returns data for each shard

        Parameters
        ----------
        input_files: List[str]
            List of file names to process
        n_samples:  int, optional
            The number of samples to extract from each variant in the data
        shard_size: int, optional (default 4096)
            The size of a shard of data to process at a time. Here, shard_size is equal to the
            number of variants to fetch. You can think of them as number of rows to get from the
            full dataset.

        Yields
        -------
        Iterator
            Iterator over shards
        """

        return _fastq_load_files(input_files, shard_size)

    def create_dataset(self,
                       input_files: OneOrMany[str],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 4096) -> DiskDataset:
        """Creates a `Dataset` from input FASTQ files.

        Parameters
        ----------
        input_files: List[str]
            List of fastQ files.
        data_dir: str, optional (default None)
            Name of directory where featurized data is stored.
          shard_size: int, optional (default 4096)

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation of data
            from `input_files`.
        """
        if isinstance(input_files, str):
            input_files = [input_files]

        def shard_generator():

            for shard_num, shard in enumerate(
                    self._get_shards(input_files, shard_size)):
                if self.return_quality_scores:
                    sequences, quality_scores = _generate_sequences(shard)
                    # Featurize sequences
                    X = self.featurizer(sequences)
                    ids = np.ones(len(X))
                    # (X, y , w, ids)
                    yield X, None, quality_scores, ids
                else:
                    sequences = _generate_sequences(shard)
                    # Featurize sequences
                    X = self.featurizer(sequences)
                    ids = np.ones(len(X))
                    # (X, y , w, ids)
                    yield X, None, None, ids

        def _generate_sequences(shard: List) -> OneOrMany[np.ndarray]:
            """
            Creates a numpy array of annotated FASTQ-format strings.
            """
            assert len(
                shard
            ) % 4 == 0, f'Sharded length not divisible by four: Length of shard = {len(shard)}. File is possibly incomplete'
            sequences: np.ndarray = np.array([], dtype='object')
            if self.return_quality_scores:
                quality_scores: np.ndarray = np.array([], dtype='object')

            # Go through each sequence entity in the fastq_file: each sequence consists of 4 lines
            # First line : header description
            # second line : sequence
            # third line : more description usually the same as the first line
            # fourth line: quality scores of the sequence
            for start_index in range(0, len(shard), 4):
                each_sequence = shard[start_index:start_index + 4]

                # Second line : add sequence to the sequence array
                sequences = _add_sequence(
                    sequences, np.array([each_sequence[1].strip("\n")]))

                # Fourth line
                if self.return_quality_scores:
                    quality_scores = _add_sequence(
                        quality_scores,
                        np.array([each_sequence[3].strip("\n")]))

            if self.return_quality_scores:
                return sequences, quality_scores
            else:
                return sequences

        def _add_sequence(sequences: np.ndarray,
                          sequence: np.ndarray) -> np.ndarray:
            # Handle empty sequence
            if sequence is None or len(sequence) <= 0:
                return np.array([])
            # Annotate start/stop of sequence
            if self.auto_add_annotations:
                sequence = np.insert(sequence, 0, "[CLS]")
                sequence = np.append(sequence, "[SEP]")
            new_sequence = ''.join(sequence)
            new_sequences = np.append(sequences, new_sequence)
            return new_sequences

        return DiskDataset.create_dataset(shard_generator(), data_dir)


class ImageLoader(DataLoader):
    """Handles loading of image files.

    This class allows for loading of images in various formats.
    For user convenience, also accepts zip-files and directories
    of images and uses some limited intelligence to attempt to
    traverse subdirectories which contain images.
    """

    def __init__(self, tasks: Optional[List[str]] = None):
        """Initialize image loader.

        At present, custom image featurizers aren't supported by this
        loader class.

        Parameters
        ----------
        tasks: List[str], optional (default None)
            List of task names for image labels.
        """
        if tasks is None:
            tasks = []
        self.tasks = tasks

    def create_dataset(self,
                       inputs: Union[OneOrMany[str], Tuple[Any]],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 8192,
                       in_memory: bool = False) -> Dataset:
        """Creates and returns a `Dataset` object by featurizing provided image files and labels/weights.

        Parameters
        ----------
        inputs: `Union[OneOrMany[str], Tuple[Any]]`
            The inputs provided should be one of the following

              - filename
              - list of filenames
              - Tuple (list of filenames, labels)
              - Tuple (list of filenames, labels, weights)

            Each file in a given list of filenames should either be of a supported
            image format (.png, .tif only for now) or of a compressed folder of
            image files (only .zip for now). If `labels` or `weights` are provided,
            they must correspond to the sorted order of all filenames provided, with
            one label/weight per file.
        data_dir: str, optional (default None)
            Directory to store featurized dataset.
        shard_size: int, optional (default 8192)
            Shard size when loading data.
        in_memory: bool, optioanl (default False)
            If true, return in-memory NumpyDataset. Else return ImageDataset.

        Returns
        -------
        ImageDataset or NumpyDataset or DiskDataset
            - if `in_memory == False`, the return value is ImageDataset.
            - if `in_memory == True` and `data_dir is None`, the return value is NumpyDataset.
            - if `in_memory == True` and `data_dir is not None`, the return value is DiskDataset.
        """
        labels, weights = None, None
        if isinstance(inputs, tuple):
            if len(inputs) == 1:
                input_files = inputs[0]
                if isinstance(inputs, str):
                    input_files = [inputs]
            elif len(inputs) == 2:
                input_files, labels = inputs
            elif len(inputs) == 3:
                input_files, labels, weights = inputs
            else:
                raise ValueError("Input must be a tuple of length 1, 2, or 3")
        else:
            input_files = inputs
        if isinstance(input_files, str):
            input_files = [input_files]

        image_files = []
        # Sometimes zip files contain directories within. Traverse directories
        while len(input_files) > 0:
            remainder = []
            for input_file in input_files:
                filename, extension = os.path.splitext(input_file)
                extension = extension.lower()
                # TODO(rbharath): Add support for more extensions
                if os.path.isdir(input_file):
                    dirfiles = [
                        os.path.join(input_file, subfile)
                        for subfile in os.listdir(input_file)
                    ]
                    remainder += dirfiles
                elif extension == ".zip":
                    zip_dir = tempfile.mkdtemp()
                    zip_ref = zipfile.ZipFile(input_file, 'r')
                    zip_ref.extractall(path=zip_dir)
                    zip_ref.close()
                    zip_files = [
                        os.path.join(zip_dir, name)
                        for name in zip_ref.namelist()
                    ]
                    for zip_file in zip_files:
                        _, extension = os.path.splitext(zip_file)
                        extension = extension.lower()
                        if extension in [".png", ".tif"]:
                            image_files.append(zip_file)
                elif extension in [".png", ".tif"]:
                    image_files.append(input_file)
                else:
                    raise ValueError("Unsupported file format")
            input_files = remainder

        # Sort image files
        image_files = sorted(image_files)

        if in_memory:
            if data_dir is None:
                return NumpyDataset(load_image_files(image_files),
                                    y=labels,
                                    w=weights,
                                    ids=image_files)
            else:
                dataset = DiskDataset.from_numpy(load_image_files(image_files),
                                                 y=labels,
                                                 w=weights,
                                                 ids=image_files,
                                                 tasks=self.tasks,
                                                 data_dir=data_dir)
                if shard_size is not None:
                    dataset.reshard(shard_size)
                return dataset
        else:
            return ImageDataset(image_files,
                                y=labels,
                                w=weights,
                                ids=image_files)


class InMemoryLoader(DataLoader):
    """Facilitate Featurization of In-memory objects.

    When featurizing a dataset, it's often the case that the initial set of
    data (pre-featurization) fits handily within memory. (For example, perhaps
    it fits within a column of a pandas DataFrame.) In this case, it would be
    convenient to directly be able to featurize this column of data. However,
    the process of featurization often generates large arrays which quickly eat
    up available memory. This class provides convenient capabilities to process
    such in-memory data by checkpointing generated features periodically to
    disk.

    Example
    -------
    Here's an example with only datapoints and no labels or weights.

    >>> import deepchem as dc
    >>> smiles = ["C", "CC", "CCC", "CCCC"]
    >>> featurizer = dc.feat.CircularFingerprint()
    >>> loader = dc.data.InMemoryLoader(tasks=["task1"], featurizer=featurizer)
    >>> dataset = loader.create_dataset(smiles, shard_size=2)
    >>> len(dataset)
    4

    Here's an example with both datapoints and labels

    >>> import deepchem as dc
    >>> smiles = ["C", "CC", "CCC", "CCCC"]
    >>> labels = [1, 0, 1, 0]
    >>> featurizer = dc.feat.CircularFingerprint()
    >>> loader = dc.data.InMemoryLoader(tasks=["task1"], featurizer=featurizer)
    >>> dataset = loader.create_dataset(zip(smiles, labels), shard_size=2)
    >>> len(dataset)
    4

    Here's an example with datapoints, labels, weights and ids all provided.

    >>> import deepchem as dc
    >>> smiles = ["C", "CC", "CCC", "CCCC"]
    >>> labels = [1, 0, 1, 0]
    >>> weights = [1.5, 0, 1.5, 0]
    >>> ids = ["C", "CC", "CCC", "CCCC"]
    >>> featurizer = dc.feat.CircularFingerprint()
    >>> loader = dc.data.InMemoryLoader(tasks=["task1"], featurizer=featurizer)
    >>> dataset = loader.create_dataset(zip(smiles, labels, weights, ids), shard_size=2)
    >>> len(dataset)
    4

    """

    def create_dataset(self,
                       inputs: Sequence[Any],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 8192) -> DiskDataset:
        """Creates and returns a `Dataset` object by featurizing provided files.

        Reads in `inputs` and uses `self.featurizer` to featurize the
        data in these input files.  For large files, automatically shards
        into smaller chunks of `shard_size` datapoints for convenience.
        Returns a `Dataset` object that contains the featurized dataset.

        This implementation assumes that the helper methods `_get_shards`
        and `_featurize_shard` are implemented and that each shard
        returned by `_get_shards` is a pandas dataframe.  You may choose
        to reuse or override this method in your subclass implementations.

        Parameters
        ----------
        inputs: Sequence[Any]
            List of inputs to process. Entries can be arbitrary objects so long as
            they are understood by `self.featurizer`
        data_dir: str, optional (default None)
            Directory to store featurized dataset.
        shard_size: int, optional (default 8192)
            Number of examples stored in each shard.

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation of data
            from `inputs`.
        """
        logger.info("Loading raw samples now.")
        logger.info("shard_size: %s" % str(shard_size))

        if not isinstance(inputs, list):
            try:
                inputs = list(inputs)
            except TypeError:
                inputs = [inputs]

        def shard_generator():
            global_index = 0
            for shard_num, shard in enumerate(
                    self._get_shards(inputs, shard_size)):
                time1 = time.time()
                X, y, w, ids = self._featurize_shard(shard, global_index)
                global_index += len(shard)

                time2 = time.time()
                logger.info("TIMING: featurizing shard %d took %0.3f s" %
                            (shard_num, time2 - time1))
                yield X, y, w, ids

        return DiskDataset.create_dataset(shard_generator(), data_dir,
                                          self.tasks)

    def _get_shards(self, inputs: List,
                    shard_size: Optional[int]) -> Iterator[pd.DataFrame]:
        """Break up input into shards.

        Parameters
        ----------
        inputs: List
            Each entry in this list must be of the form `(featurization_input,
            label, weight, id)` or `(featurization_input, label, weight)` or
            `(featurization_input, label)` or `featurization_input` for one
            datapoint, where `featurization_input` is any input that is recognized
            by `self.featurizer`.
        shard_size: int, optional
            The size of shard to generate.

        Returns
        -------
        Iterator[pd.DataFrame]
            Iterator which iterates over shards of data.
        """
        current_shard: List = []
        for i, datapoint in enumerate(inputs):
            if i != 0 and shard_size is not None and i % shard_size == 0:
                shard_data = current_shard
                current_shard = []
                yield shard_data
            current_shard.append(datapoint)
        yield current_shard

    # FIXME: Signature of "_featurize_shard" incompatible with supertype "DataLoader"
    def _featurize_shard(  # type: ignore[override]
        self, shard: List, global_index: int) -> Tuple[np.ndarray, np.ndarray,
                                                       np.ndarray, np.ndarray]:
        """Featurizes a shard of an input data.

        Parameters
        ----------
        shard: List
            List each entry of which must be of the form `(featurization_input,
            label, weight, id)` or `(featurization_input, label, weight)` or
            `(featurization_input, label)` or `featurization_input` for one
            datapoint, where `featurization_input` is any input that is recognized
            by `self.featurizer`.
        global_index: int
            The starting index for this shard in the full set of provided inputs

        Returns
        ------
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
            The tuple is `(X, y, w, ids)`. All values are numpy arrays.
        """
        features = []
        labels = []
        weights = []
        ids = []
        n_tasks = len(self.tasks)
        for i, entry in enumerate(shard):
            if not isinstance(entry, tuple):
                entry = (entry,)
            if len(entry) > 4:
                raise ValueError(
                    "Entry is malformed and must be of length 1-4 containing featurization_input"
                    "and optionally label, weight, and id.")
            if len(entry) == 4:
                featurization_input, label, weight, entry_id = entry
            elif len(entry) == 3:
                featurization_input, label, weight = entry
                entry_id = global_index + i
            elif len(entry) == 2:
                featurization_input, label = entry
                weight = np.ones((n_tasks), np.float32)
                entry_id = global_index + i
            elif len(entry) == 1:
                featurization_input = entry
                label = np.zeros((n_tasks), np.float32)
                weight = np.zeros((n_tasks), np.float32)
                entry_id = global_index + i
            feature = self.featurizer(featurization_input)
            features.append(feature)
            weights.append(weight)
            labels.append(label)
            ids.append(entry_id)
        X = np.concatenate(features, axis=0)
        return X, np.array(labels), np.array(weights), np.array(ids)


class DFTYamlLoader(DataLoader):
    """
    Creates a `Dataset` object from YAML input files.

    This class provides methods to load and featurize data from a YAML file.
    Although, in this class, we only focus on a specfic input format
    that can be used to perform Density Functional Theory calculations.

    Examples
    --------
    >>> from deepchem.data.data_loader import DFTYamlLoader
    >>> import deepchem as dc
    >>> import pytest
    >>> inputs = 'deepchem/data/tests/dftdata.yaml'
    >>> data = DFTYamlLoader()
    >>> output = data.create_dataset(inputs)

    Notes
    -----
    Format (and example) for the YAML file:

    - e_type : 'ae'
      true_val : '0.09194410469'
      systems : [{
            'moldesc': 'Li 1.5070 0 0; H -1.5070 0 0',
            'basis': '6-311++G(3df,3pd)'
        }]

    Each entry in the YAML file must contain the three parameters : e_type,
    true_val and systems in this particular order.
    One entry object may contain one or more systems.
    This data class does not support/ require an additional featurizer,
    since the datapoints are featurized within the methods.
    To read more about the parameters and their possible values please refer to
    deepchem.feat.dft_data.

    """

    def __init__(self):
        """
        Initialize DFTYAML loader
        """

    def create_dataset(self,
                       inputs: OneOrMany[Any],
                       data_dir: Optional[str] = None,
                       shard_size: Optional[int] = 1) -> Dataset:
        """
        Creates and returns a `Dataset` object by featurizing provided YAML
        files.

        Parameters
        ----------
        input_files: OneOrMany[str]
            List of YAML filenames.
        data_dir: Optional[str], default None
            Name of directory where featurized data is stored.
        shard_size: int, optional (default 1)
            Shard size when loading data.

        Returns
        -------
        DiskDataset
            A `DiskDataset` object containing a featurized representation
            of data from `inputs`.
        """

        def shard_generator():
            entries = self._get_shards(inputs)
            for i, shard in enumerate(entries):
                X = np.array(self._featurize_shard(shard))
                y = X[0].get_true_val()
                w = np.array([X[0].get_weight()])
                ids = np.array([i])
                yield X, y, w, ids

        return DiskDataset.create_dataset(shard_generator(), data_dir)

    def _get_shards(self, inputs):
        """
        Loads and divides the .yaml file into shards.

        Parameters
        ----------
        input_files: str
            .yaml file to be processed.

        Returns
        -------
        data
            list of dictionaries where each dictionary corresponds to one
            shard and is then featurized into one entry object.
        """
        with open(inputs) as f:
            data = yaml.load(f, Loader=SafeLoader)
        return (data)

    def _featurize_shard(self, shard):
        """
        Featurizes shards in the dataset

        Parameters
        ----------
        shard: dict
            Dictionary containing values to initialize the DFTEntry object.

        Returns
        -------
        x: featurized shard (DFTEntry objects)
        """
        try:
            e_type = shard['e_type']
            if 'true_val' in shard.keys():
                true_val = shard['true_val']
            else:
                true_val = '0.0'
            systems = shard['systems']
        except KeyError:
            raise ValueError(
                "Unknown key in yaml file. Please check format for correctness."
            )
        if 'weight' in shard.keys():
            weight = shard['weight']
            x = DFTEntry.create(e_type, true_val, systems, weight)
        else:
            x = DFTEntry.create(e_type, true_val, systems)
        return [x]



from functools import partial
from math import sqrt
from typing import Callable, Dict, List, Union

import torch
from torch import nn

from deepchem.feat.molecule_featurizers.conformer_featurizer import (
    full_atom_feature_dims,
    full_bond_feature_dims,
)
from deepchem.models.torch_models.layers import MultilayerPerceptron
from deepchem.utils.graph_utils import (
    aggregate_max,
    aggregate_mean,
    aggregate_min,
    aggregate_moment,
    aggregate_std,
    aggregate_sum,
    aggregate_var,
    scale_amplification,
    scale_attenuation,
    scale_identity,
)

PNA_AGGREGATORS = {
    "mean": aggregate_mean,
    "sum": aggregate_sum,
    "max": aggregate_max,
    "min": aggregate_min,
    "std": aggregate_std,
    "var": aggregate_var,
    "moment3": partial(aggregate_moment, n=3),
    "moment4": partial(aggregate_moment, n=4),
    "moment5": partial(aggregate_moment, n=5),
}

PNA_SCALERS = {
    "identity": scale_identity,
    "amplification": scale_amplification,
    "attenuation": scale_attenuation,
}


class AtomEncoder(torch.nn.Module):
    """
    Encodes atom features into embeddings based on the Open Graph Benchmark feature set in conformer_featurizer.

    Parameters
    ----------
    emb_dim : int
        The dimension that the returned embedding will have.
    padding : bool, optional (default=False)
        If true then the last index will be used for padding.

    Examples
    --------
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import full_atom_feature_dims
    >>> atom_encoder = AtomEncoder(emb_dim=32)
    >>> num_rows = 10
    >>> atom_features = torch.stack([
    ... torch.randint(low=0, high=dim, size=(num_rows,))
    ... for dim in full_atom_feature_dims
    ... ], dim=1)
    >>> atom_embeddings = atom_encoder(atom_features)
    """

    def __init__(self, emb_dim, padding=False):
        super(AtomEncoder, self).__init__()

        self.atom_embedding_list = torch.nn.ModuleList()
        self.padding = padding

        for dim in full_atom_feature_dims:
            if padding:
                emb = torch.nn.Embedding(dim + 1, emb_dim, padding_idx=0)
            else:
                emb = torch.nn.Embedding(dim, emb_dim)
            torch.nn.init.xavier_uniform_(emb.weight.data)
            self.atom_embedding_list.append(emb)

    def reset_parameters(self):
        """
        Reset the parameters of the atom embeddings.

        This method resets the weights of the atom embeddings by initializing
        them with a uniform distribution between -sqrt(3) and sqrt(3).
        """
        for embedder in self.atom_embedding_list:
            embedder.weight.data.uniform_(-sqrt(3), sqrt(3))

    def forward(self, x):
        """
        Compute the atom embeddings for the given atom features.

        Parameters
        ----------
        x : torch.Tensor, shape (batch_size, num_atoms, num_features)
            The input atom features tensor.

        Returns
        -------
        x_embedding : torch.Tensor, shape (batch_size, num_atoms, emb_dim)
            The computed atom embeddings.
        """
        x_embedding = 0
        for i in range(x.shape[1]):
            if self.padding:
                x_embedding += self.atom_embedding_list[i](x[:, i].long() + 1)
            else:
                x_embedding += self.atom_embedding_list[i](x[:, i].long())

        return x_embedding


class BondEncoder(torch.nn.Module):
    """
    Encodes bond features into embeddings based on the Open Graph Benchmark feature set in conformer_featurizer.

    Parameters
    ----------
    emb_dim : int
        The dimension that the returned embedding will have.
    padding : bool, optional (default=False)
        If true then the last index will be used for padding.

    Examples
    --------
    >>> from deepchem.feat.molecule_featurizers.conformer_featurizer import full_bond_feature_dims
    >>> bond_encoder = BondEncoder(emb_dim=32)
    >>> num_rows = 10
    >>> bond_features = torch.stack([
    ... torch.randint(low=0, high=dim, size=(num_rows,))
    ... for dim in full_bond_feature_dims
    ... ], dim=1)
    >>> bond_embeddings = bond_encoder(bond_features)
    """

    def __init__(self, emb_dim, padding=False):
        super(BondEncoder, self).__init__()

        self.bond_embedding_list = torch.nn.ModuleList()
        self.padding = padding

        for dim in full_bond_feature_dims:
            if padding:
                emb = torch.nn.Embedding(dim + 1, emb_dim, padding_idx=0)
            else:
                emb = torch.nn.Embedding(dim, emb_dim)
            torch.nn.init.xavier_uniform_(emb.weight.data)
            self.bond_embedding_list.append(emb)

    def forward(self, edge_attr):
        """
        Compute the bond embeddings for the given bond features.

        Parameters
        ----------
        edge_attr : torch.Tensor, shape (batch_size, num_edges, num_features)
            The input bond features tensor.

        Returns
        -------
        bond_embedding : torch.Tensor, shape (batch_size, num_edges, emb_dim)
            The computed bond embeddings.
        """
        bond_embedding = 0
        for i in range(edge_attr.shape[1]):
            if self.padding:
                bond_embedding += self.bond_embedding_list[i](
                    edge_attr[:, i].long() + 1)
            else:
                bond_embedding += self.bond_embedding_list[i](
                    edge_attr[:, i].long())

        return bond_embedding


class PNALayer(nn.Module):
    """
    Principal Neighbourhood Aggregation Layer (PNA) from [1].

    Parameters
    ----------
    in_dim : int
        Input dimension of the node features.
    out_dim : int
        Output dimension of the node features.
    in_dim_edges : int
        Input dimension of the edge features.
    aggregators : List[str]
        List of aggregator functions to use.
    scalers : List[str]
        List of scaler functions to use.
    activation : Union[Callable, str], optional, default="relu"
        Activation function to use.
    last_activation : Union[Callable, str], optional, default="none"
        Last activation function to use.
    dropout : float, optional, default=0.0
        Dropout rate.
    residual : bool, optional, default=True
        Whether to use residual connections.
    pairwise_distances : bool, optional, default=False
        Whether to use pairwise distances.
    mid_batch_norm : bool, optional, default=False
        Whether to use batch normalization in the middle layers.
    last_batch_norm : bool, optional, default=False
        Whether to use batch normalization in the last layer.
    batch_norm_momentum : float, optional, default=0.1
        Momentum for the batch normalization layers.
    avg_d : Dict[str, float], optional, default={"log": 1.0}
        Dictionary containing the average degree of the graph.
    posttrans_layers : int, optional, default=2
        Number of post-transformation layers.
    pretrans_layers : int, optional, default=1
        Number of pre-transformation layers.

    References
    ----------
    .. [1] Corso, G., Cavalleri, L., Beaini, D., Liò, P. & Veličković, P. Principal Neighbourhood Aggregation for Graph Nets. Preprint at https://doi.org/10.48550/arXiv.2004.05718 (2020).

    Examples
    --------
    >>> import dgl
    >>> import numpy as np
    >>> import torch
    >>> from deepchem.models.torch_models.pna_gnn import PNALayer
    >>> in_dim = 32
    >>> out_dim = 64
    >>> in_dim_edges = 16
    >>> aggregators = ["mean", "max"]
    >>> scalers = ["identity", "amplification", "attenuation"]
    >>> pna_layer = PNALayer(in_dim=in_dim,
    ...                      out_dim=out_dim,
    ...                      in_dim_edges=in_dim_edges,
    ...                      aggregators=aggregators,
    ...                      scalers=scalers)
    >>> num_nodes = 10
    >>> num_edges = 20
    >>> node_features = torch.randn(num_nodes, in_dim)
    >>> edge_features = torch.randn(num_edges, in_dim_edges)
    >>> g = dgl.graph((np.random.randint(0, num_nodes, num_edges),
    ...                np.random.randint(0, num_nodes, num_edges)))
    >>> g.ndata['feat'] = node_features
    >>> g.edata['feat'] = edge_features
    >>> g.ndata['feat'] = pna_layer(g)
    """

    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        in_dim_edges: int,
        aggregators: List[str],
        scalers: List[str],
        activation: Union[Callable, str] = "relu",
        dropout: float = 0.0,
        residual: bool = True,
        pairwise_distances: bool = False,
        batch_norm_momentum=0.1,
        avg_d: Dict[str, float] = {"log": 1.0},
        posttrans_layers: int = 2,
        pretrans_layers: int = 1,
    ):
        super(PNALayer, self).__init__()
        self.aggregators = [PNA_AGGREGATORS[aggr] for aggr in aggregators]
        self.scalers = [PNA_SCALERS[scale] for scale in scalers]
        self.edge_features = in_dim_edges > 0
        self.activation = activation
        self.avg_d = avg_d
        self.pairwise_distances = pairwise_distances
        self.residual = residual
        if in_dim != out_dim:
            self.residual = False

        self.pretrans = MultilayerPerceptron(
            d_input=(2 * in_dim + in_dim_edges +
                     1) if self.pairwise_distances else
            (2 * in_dim + in_dim_edges),
            d_output=in_dim,
            d_hidden=(in_dim,) * (pretrans_layers - 1),
            batch_norm=True,
            batch_norm_momentum=batch_norm_momentum,
            dropout=dropout)

        self.posttrans = MultilayerPerceptron(
            d_input=(len(self.aggregators) * len(self.scalers) + 1) * in_dim,
            d_hidden=(out_dim,) * (posttrans_layers - 1),
            d_output=out_dim,
            batch_norm=True,
            batch_norm_momentum=batch_norm_momentum,
            dropout=dropout)

    def forward(self, g):
        """
        Forward pass of the PNA layer.

        Parameters
        ----------
        g : dgl.DGLGraph
            Input graph

        Returns
        -------
        h : torch.Tensor
            Node feature tensor
        """
        h = g.ndata['feat']
        h_in = h
        # pretransformation
        g.apply_edges(self.pretrans_edges)

        # aggregation
        g.update_all(self.message_func, self.reduce_func)
        h = torch.cat([h, g.ndata['feat']], dim=-1)
        # post-transformation
        h = self.posttrans(h)
        if self.residual:
            h = h + h_in

        return h

    def message_func(self, edges) -> Dict[str, torch.Tensor]:
        """
        The message function to generate messages along the edges.

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Batch of edges.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary containing the edge features.
        """
        return {"e": edges.data["e"]}

    def reduce_func(self, nodes) -> Dict[str, torch.Tensor]:
        """
        The reduce function to aggregate the messages.
        Apply the aggregators and scalers, and concatenate the results.

        Parameters
        ----------
        nodes : dgl.NodeBatch
            Batch of nodes.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary containing the aggregated node features.
        """
        h_in = nodes.data['feat']
        h = nodes.mailbox["e"]
        D = h.shape[-2]

        h_to_cat = [
            aggr(h=h, h_in=h_in)  # type: ignore
            for aggr in self.aggregators
        ]
        h = torch.cat(h_to_cat, dim=-1)

        if len(self.scalers) > 1:
            h = torch.cat(
                [
                    scale(h, D=D, avg_d=self.avg_d)  # type: ignore
                    for scale in self.scalers
                ],
                dim=-1)

        return {'feat': h}

    def pretrans_edges(self, edges) -> Dict[str, torch.Tensor]:
        """
        Return a mapping to the concatenation of the features from
        the source node, the destination node, and the edge between them (if applicable).

        Parameters
        ----------
        edges : dgl.EdgeBatch
            Batch of edges.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dictionary containing the concatenated features.
        """

        if self.edge_features and self.pairwise_distances:
            squared_distance = torch.sum((edges.src['x'] - edges.dst['x'])**2,
                                         dim=-1)[:, None]
            z2 = torch.cat([
                edges.src['feat'], edges.dst['feat'], edges.data['feat'],
                squared_distance
            ],
                           dim=-1)
        elif not self.edge_features and self.pairwise_distances:
            squared_distance = torch.sum((edges.src['x'] - edges.dst['x'])**2,
                                         dim=-1)[:, None]
            z2 = torch.cat(
                [edges.src['feat'], edges.dst['feat'], squared_distance],
                dim=-1)
        elif self.edge_features and not self.pairwise_distances:
            z2 = torch.cat(
                [edges.src['feat'], edges.dst['feat'], edges.data['feat']],
                dim=-1)
        else:
            z2 = torch.cat([edges.src['feat'], edges.dst['feat']], dim=-1)
        return {"e": self.pretrans(z2)}
